#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import sys

from PyQt5.QtCore import QFile, QSize, QTextStream, Qt
from PyQt5.QtWidgets import QDialog, QDialogButtonBox, QLabel, QScrollArea, QStyleFactory, QVBoxLayout, qApp


class Changelog(QDialog):
    def __init__(self, parent=None):
        super(Changelog, self).__init__(parent, Qt.Dialog | Qt.WindowCloseButtonHint)
        self.parent = parent
        self.setWindowTitle('{} changelog'.format(qApp.applicationName()))
        changelog = QFile(':/CHANGELOG')
        changelog.open(QFile.ReadOnly | QFile.Text)
        content = QTextStream(changelog).readAll()
        label = QLabel(content, self)
        label.setWordWrap(True)
        label.setTextFormat(Qt.PlainText)
        buttons = QDialogButtonBox(QDialogButtonBox.Close, self)
        buttons.rejected.connect(self.close)
        scrollarea = QScrollArea(self)
        scrollarea.setStyleSheet('QScrollArea { background:transparent; }')
        scrollarea.setWidgetResizable(True)
        scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scrollarea.setFrameShape(QScrollArea.NoFrame)
        scrollarea.setWidget(label)
        if sys.platform in {'win32', 'darwin'}:
            scrollarea.setStyle(QStyleFactory.create('Fusion'))
        # noinspection PyUnresolvedReferences
        if parent.parent.stylename == 'fusion' or sys.platform in {'win32', 'darwin'}:
            self.setStyleSheet('''
            QScrollArea {{
                background-color: transparent;
                margin-bottom: 10px;
                border: none;
                border-right: 1px solid {};
            }}'''.format('#4D5355' if parent.theme == 'dark' else '#C0C2C3'))
        else:
            self.setStyleSheet('''
            QScrollArea {{
                background-color: transparent;
                margin-bottom: 10px;
                border: none;
            }}''')
        layout = QVBoxLayout()
        layout.addWidget(scrollarea)
        layout.addWidget(buttons)
        self.setLayout(layout)
        self.setMinimumSize(self.sizeHint())

    def sizeHint(self) -> QSize:
        modes = {
            'LOW': QSize(450, 300),
            'NORMAL': QSize(565, 560),
            'HIGH': QSize(1080, 920)
        }
        return modes[self.parent.parentWidget().scale]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import os
import sys

from PyQt5.QtCore import pyqtSlot, Qt
from PyQt5.QtGui import QCloseEvent, QMouseEvent, QPixmap
from PyQt5.QtWidgets import (QCheckBox, QDialog, QDialogButtonBox, QFrame, QGridLayout, QGroupBox, QHBoxLayout, QLabel,
                             QMessageBox, QScrollArea, QSizePolicy, QSpacerItem, QStyleFactory, QVBoxLayout, QWidget)

from vidcutter.libs.iso639 import ISO639_2
from vidcutter.libs.videoservice import VideoService


class StreamSelector(QDialog):
    def __init__(self, service: VideoService, parent=None, flags=Qt.Dialog | Qt.WindowCloseButtonHint):
        super(StreamSelector, self).__init__(parent, flags)
        self.service = service
        self.parent = parent
        self.streams = service.streams
        self.config = service.mappings
        self.setObjectName('streamselector')
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle('Media streams - {}'.format(os.path.basename(self.parent.currentMedia)))
        buttons = QDialogButtonBox(QDialogButtonBox.Ok, self)
        buttons.accepted.connect(self.close)
        layout = QVBoxLayout()
        layout.setSpacing(15)
        if len(self.streams.video):
            layout.addWidget(self.video())
        if len(self.streams.audio):
            layout.addWidget(self.audio())
        if len(self.streams.subtitle):
            layout.addWidget(self.subtitles())
        layout.addWidget(buttons)
        self.setLayout(layout)

    @staticmethod
    def lineSeparator() -> QFrame:
        line = QFrame()
        line.setFrameShape(QFrame.HLine)
        line.setFrameShadow(QFrame.Sunken)
        line.setLineWidth(1)
        line.setMidLineWidth(0)
        line.setMinimumSize(0, 2)
        return line

    def video(self) -> QGroupBox:
        framerate = round(eval(self.streams.video.avg_frame_rate), 3)
        ratio = self.streams.video.display_aspect_ratio.split(':')
        ratio = round(int(ratio[0]) / int(ratio[1]), 3)
        icon = QLabel('<img src=":images/{}/streams-video.png" />'.format(self.parent.theme), self)
        label = QLabel('''
            <b>index:</b> {index}
            <br/>
            <b>codec:</b> {codec}
            <br/>
            <b>size:</b> {width} x {height}
            &nbsp;
            <b>ratio:</b> {ratio}
            <br/>
            <b>frame rate:</b> {framerate} fps
            &nbsp;
            <b>color format:</b> {pixfmt}'''.format(index=self.streams.video.index,
                                                    codec=self.streams.video.codec_long_name,
                                                    width=self.streams.video.width,
                                                    height=self.streams.video.height,
                                                    framerate='{0:.2f}'.format(framerate),
                                                    ratio='{0:.2f}'.format(ratio),
                                                    pixfmt=self.streams.video.pix_fmt), self)
        label.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        videolayout = QHBoxLayout()
        videolayout.setSpacing(15)
        videolayout.addSpacing(25)
        videolayout.addWidget(icon)
        videolayout.addSpacing(45)
        videolayout.addWidget(label)
        videogroup = QGroupBox('Video')
        videogroup.setLayout(videolayout)
        return videogroup

    def audio(self) -> QGroupBox:
        audiolayout = QGridLayout()
        audiolayout.setSpacing(15)
        for stream in self.streams.audio:
            sameplerate = round(int(stream.sample_rate) / 1000, 1)
            checkbox = StreamSelectorCheckBox(stream.index, 'Toggle audio stream', self)
            icon = StreamSelectorLabel('<img src=":images/{}/streams-audio.png" />'.format(self.parent.theme),
                                       checkbox, True, self)
            labeltext = '<b>index:</b> {}<br/>'.format(stream.index)
            if hasattr(stream, 'tags') and hasattr(stream.tags, 'language'):
                labeltext += '<b>language:</b> {}<br/>'.format(ISO639_2[stream.tags.language])
            labeltext += '<b>codec:</b> {}<br/>'.format(stream.codec_long_name)
            labeltext += '<b>channels:</b> {0} &nbsp; <b>sample rate:</b> {1:.2f} kHz' \
                         .format(stream.channels, sameplerate)
            label = StreamSelectorLabel(labeltext, checkbox, False, self)
            rows = audiolayout.rowCount()
            audiolayout.addWidget(checkbox, rows, 0)
            audiolayout.addItem(QSpacerItem(15, 1), rows, 1)
            audiolayout.addWidget(icon, rows, 2)
            audiolayout.addItem(QSpacerItem(30, 1), rows, 3)
            audiolayout.addWidget(label, rows, 4)
            if self.streams.audio.index(stream) < len(self.streams.audio) - 1:
                audiolayout.addWidget(StreamSelector.lineSeparator(), rows + 1, 0, 1, 5)
        audiolayout.setColumnStretch(4, 1)
        audiogroup = QGroupBox('Audio')
        if len(self.streams.audio) > 2:
            audiolayout.setSizeConstraint(QGridLayout.SetMinAndMaxSize)
            widget = QWidget(self)
            widget.setObjectName('audiowidget')
            widget.setStyleSheet('QWidget#audiowidget { background-color: transparent; }')
            widget.setMinimumWidth(400)
            widget.setLayout(audiolayout)
            scrolllayout = QHBoxLayout()
            scrolllayout.addWidget(StreamSelectorScrollArea(widget, 200, self.parent.theme, self))
            audiogroup.setLayout(scrolllayout)
        else:
            audiogroup.setLayout(audiolayout)
        return audiogroup

    def subtitles(self) -> QGroupBox:
        subtitlelayout = QGridLayout()
        subtitlelayout.setSpacing(15)
        for stream in self.streams.subtitle:
            checkbox = StreamSelectorCheckBox(stream.index, 'Toggle subtitle stream', self)
            icon = StreamSelectorLabel('<img src=":images/{}/streams-subtitle.png" />'.format(self.parent.theme),
                                       checkbox, True, self)
            labeltext = '<b>index:</b> {}<br/>'.format(stream.index)
            if hasattr(stream, 'tags') and hasattr(stream.tags, 'language'):
                labeltext += '<b>language:</b> {}<br/>'.format(ISO639_2[stream.tags.language])
            labeltext += '<b>codec:</b> {}'.format(stream.codec_long_name)
            label = StreamSelectorLabel(labeltext, checkbox, False, self)
            rows = subtitlelayout.rowCount()
            subtitlelayout.addWidget(checkbox, rows, 0)
            subtitlelayout.addItem(QSpacerItem(15, 1), rows, 1)
            subtitlelayout.addWidget(icon, rows, 2)
            subtitlelayout.addItem(QSpacerItem(30, 1), rows, 3)
            subtitlelayout.addWidget(label, rows, 4)
            if self.streams.subtitle.index(stream) < len(self.streams.subtitle) - 1:
                subtitlelayout.addWidget(StreamSelector.lineSeparator(), rows + 1, 0, 1, 5)
        subtitlelayout.setColumnStretch(4, 1)
        subtitlegroup = QGroupBox('Subtitles')
        if len(self.streams.subtitle) > 2:
            subtitlelayout.setSizeConstraint(QVBoxLayout.SetMinAndMaxSize)
            widget = QWidget(self)
            widget.setObjectName('subtitlewidget')
            widget.setStyleSheet('QWidget#subtitlewidget { background-color: transparent; }')
            widget.setMinimumWidth(400)
            widget.setLayout(subtitlelayout)
            scrolllayout = QHBoxLayout()
            scrolllayout.addWidget(StreamSelectorScrollArea(widget, 170, self.parent.theme, self))
            subtitlegroup.setStyleSheet('QGroupBox { padding-right: 0; }')
            subtitlegroup.setLayout(scrolllayout)
        else:
            subtitlegroup.setLayout(subtitlelayout)
        return subtitlegroup

    @pyqtSlot()
    def closeEvent(self, event: QCloseEvent) -> None:
        # check if all audio streams are off
        idx = [stream.index for stream in self.streams.audio]
        no_audio = len(self.streams.audio) and True not in [self.config[i] for i in idx]
        # check if all subtitle streams are off
        idx = [stream.index for stream in self.streams.subtitle]
        no_subtitles = len(self.streams.subtitle) and True not in [self.config[i] for i in idx]
        # warn user if all audio and/or subtitle streams are off
        if no_audio or no_subtitles:
            if no_audio and not no_subtitles:
                warnsubtext = 'All audio streams have been deselected which will produce a file with <b>NO AUDIO</b> ' \
                              'when you save.'
            elif not no_audio and no_subtitles:
                warnsubtext = 'All subtitle streams have been deselected which will produce a file with ' \
                              '<b>NO SUBTITLES</b> when you save.'
            else:
                warnsubtext = 'All audio and subtitle streams have been deselected which will produce a file ' \
                              'with <b>NO AUDIO</b> and <b>NO SUBTITLES</b> when you save.'
            warntext = '''
                <style>
                    h2 {{
                        color: {};
                        font-family: "Futura LT", sans-serif;
                        font-weight: normal;
                    }}
                </style>
                <table border="0" cellpadding="6" cellspacing="0" width="350">
                    <tr>
                        <td><h2>A friendly configuration warning</h2></td>
                    </tr>
                    <tr>
                        <td>{}</td>
                    </tr>
                    <tr>
                        <td>Are you sure this is what you want?</td>
                    </tr>
                </table>'''.format('#C681D5' if self.parent.theme == 'dark' else '#642C68', warnsubtext)
            warnmsg = QMessageBox(QMessageBox.Warning, 'Warning', warntext, parent=self)
            warnmsg.setIconPixmap(QPixmap(':images/warning.png'))
            warnmsg.addButton('Yes', QMessageBox.YesRole)
            cancelbtn = warnmsg.addButton('No', QMessageBox.RejectRole)
            warnmsg.exec_()
            res = warnmsg.clickedButton()
            if res == cancelbtn:
                event.ignore()
                return
        event.accept()
        self.deleteLater()
        super(StreamSelector, self).closeEvent(event)


class StreamSelectorScrollArea(QScrollArea):
    def __init__(self, widget: QWidget, minHeight: int, theme: str, parent):
        super(StreamSelectorScrollArea, self).__init__(parent)
        if sys.platform in {'win32', 'darwin'}:
            self.setStyle(QStyleFactory.create('Fusion'))
        # noinspection PyUnresolvedReferences
        if parent.parent.parent.stylename == 'fusion' or sys.platform in {'win32', 'darwin'}:
            self.setStyleSheet('''
            QScrollArea {{
                background-color: transparent;
                margin-bottom: 10px;
                border: none;
                border-right: 1px solid {};
            }}'''.format('#4D5355' if theme == 'dark' else '#C0C2C3'))
        else:
            self.setStyleSheet('''
            QScrollArea {{
                background-color: transparent;
                margin-bottom: 10px;
                border: none;
            }}''')
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        self.setFrameShape(QFrame.NoFrame)
        self.setMinimumHeight(minHeight)
        if widget is not None:
            self.setWidget(widget)


class StreamSelectorCheckBox(QCheckBox):
    def __init__(self, stream_index: int, tooltip: str, parent):
        super(StreamSelectorCheckBox, self).__init__(parent)
        self.parent = parent
        self.setObjectName('streamcheckbox')
        self.setCursor(Qt.PointingHandCursor)
        self.setToolTip(tooltip)
        self.setChecked(self.parent.config[stream_index])
        self.stateChanged.connect(lambda state, index=stream_index: self.updateConfig(index, state == Qt.Checked))

    def updateConfig(self, index: int, checked: bool) -> None:
        self.parent.config[index] = checked


class StreamSelectorLabel(QLabel):
    def __init__(self, text: str, checkbox: StreamSelectorCheckBox, is_icon: bool=False, parent=None):
        super(StreamSelectorLabel, self).__init__(parent)
        self.checkbox = checkbox
        self.setAttribute(Qt.WA_Hover, True)
        self.setText(text)
        self.setToolTip(self.checkbox.toolTip())
        self.setCursor(Qt.PointingHandCursor)
        if is_icon:
            self.setFixedSize(18, 18)
            self.setSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        else:
            self.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Fixed)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        if event.button() == Qt.LeftButton and self.checkbox is not None:
            self.checkbox.toggle()
        super(StreamSelectorLabel, self).mousePressEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import logging
import os
import re
import sys
import time
from datetime import timedelta
from functools import partial
from typing import Callable, List, Optional, Union

from PyQt5.QtCore import (pyqtSignal, pyqtSlot, QBuffer, QByteArray, QDir, QFile, QFileInfo, QModelIndex, QPoint, QSize,
                          Qt, QTextStream, QTime, QTimer, QUrl)
from PyQt5.QtGui import QDesktopServices, QFont, QFontDatabase, QIcon, QKeyEvent, QPixmap, QShowEvent
from PyQt5.QtWidgets import (QAction, qApp, QApplication, QDialog, QFileDialog, QFrame, QGroupBox, QHBoxLayout, QLabel,
                             QListWidgetItem, QMainWindow, QMenu, QMessageBox, QPushButton, QSizePolicy, QStyleFactory,
                             QVBoxLayout, QWidget)

import sip

# noinspection PyUnresolvedReferences
from vidcutter import resources
from vidcutter.about import About
from vidcutter.changelog import Changelog
from vidcutter.mediainfo import MediaInfo
from vidcutter.mediastream import StreamSelector
from vidcutter.settings import SettingsDialog
from vidcutter.updater import Updater
from vidcutter.videolist import VideoList
from vidcutter.videoslider import VideoSlider
from vidcutter.videosliderwidget import VideoSliderWidget
from vidcutter.videostyle import VideoStyleDark, VideoStyleLight

from vidcutter.libs.config import Config, InvalidMediaException, VideoFilter
from vidcutter.libs.mpvwidget import mpvWidget
from vidcutter.libs.munch import Munch
from vidcutter.libs.notifications import JobCompleteNotification
from vidcutter.libs.taskbarprogress import TaskbarProgress
from vidcutter.libs.videoservice import VideoService
from vidcutter.libs.widgets import (ClipErrorsDialog, VCBlinkText, VCDoubleInputDialog, VCFilterMenuAction,
                                    VCFrameCounter, VCInputDialog, VCMessageBox, VCProgressDialog, VCTimeCounter,
                                    VCToolBarButton, VCVolumeSlider)

import vidcutter


class VideoCutter(QWidget):
    errorOccurred = pyqtSignal(str)

    timeformat = 'hh:mm:ss.zzz'
    runtimeformat = 'hh:mm:ss'

    def __init__(self, parent: QMainWindow):
        super(VideoCutter, self).__init__(parent)
        self.setObjectName('videocutter')
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.theme = self.parent.theme
        self.workFolder = self.parent.WORKING_FOLDER
        self.settings = self.parent.settings
        self.filter_settings = Config.filter_settings()
        self.currentMedia, self.mediaAvailable, self.mpvError = None, False, False
        self.projectDirty, self.projectSaved, self.debugonstart = False, False, False
        self.smartcut_monitor, self.notify = None, None
        self.fonts = []

        self.initTheme()
        self.updater = Updater(self.parent)

        self.seekSlider = VideoSlider(self)
        self.seekSlider.sliderMoved.connect(self.setPosition)
        self.sliderWidget = VideoSliderWidget(self, self.seekSlider)
        self.sliderWidget.setLoader(True)

        self.taskbar = TaskbarProgress(self.parent)

        self.clipTimes = []
        self.inCut, self.newproject = False, False
        self.finalFilename = ''
        self.totalRuntime, self.frameRate = 0, 0
        self.notifyInterval = 1000

        self.createChapters = self.settings.value('chapters', 'on', type=str) in {'on', 'true'}
        self.enableOSD = self.settings.value('enableOSD', 'on', type=str) in {'on', 'true'}
        self.hardwareDecoding = self.settings.value('hwdec', 'on', type=str) in {'on', 'auto'}
        self.enablePBO = self.settings.value('enablePBO', 'off', type=str) in {'on', 'true'}
        self.keepRatio = self.settings.value('aspectRatio', 'keep', type=str) == 'keep'
        self.keepClips = self.settings.value('keepClips', 'off', type=str) in {'on', 'true'}
        self.nativeDialogs = self.settings.value('nativeDialogs', 'on', type=str) in {'on', 'true'}
        self.indexLayout = self.settings.value('indexLayout', 'right', type=str)
        self.timelineThumbs = self.settings.value('timelineThumbs', 'on', type=str) in {'on', 'true'}
        self.showConsole = self.settings.value('showConsole', 'off', type=str) in {'on', 'true'}
        self.smartcut = self.settings.value('smartcut', 'off', type=str) in {'on', 'true'}
        self.level1Seek = self.settings.value('level1Seek', 2, type=float)
        self.level2Seek = self.settings.value('level2Seek', 5, type=float)
        self.verboseLogs = self.parent.verboseLogs
        self.lastFolder = self.settings.value('lastFolder', QDir.homePath(), type=str)

        self.videoService = VideoService(self.settings, self)
        self.videoService.progress.connect(self.seekSlider.updateProgress)
        self.videoService.finished.connect(self.smartmonitor)
        self.videoService.error.connect(self.completeOnError)
        self.videoService.addScenes.connect(self.addScenes)

        self.project_files = {
            'edl': re.compile(r'(\d+(?:\.?\d+)?)\t(\d+(?:\.?\d+)?)\t([01])'),
            'vcp': re.compile(r'(\d+(?:\.?\d+)?)\t(\d+(?:\.?\d+)?)\t([01])\t(".*")$')
        }

        self._initIcons()
        self._initActions()

        self.appmenu = QMenu(self.parent)
        self.clipindex_removemenu, self.clipindex_contextmenu = QMenu(self), QMenu(self)

        self._initMenus()
        self._initNoVideo()

        self.cliplist = VideoList(self)
        self.cliplist.customContextMenuRequested.connect(self.itemMenu)
        self.cliplist.currentItemChanged.connect(self.selectClip)
        self.cliplist.model().rowsInserted.connect(self.setProjectDirty)
        self.cliplist.model().rowsRemoved.connect(self.setProjectDirty)
        self.cliplist.model().rowsMoved.connect(self.setProjectDirty)
        self.cliplist.model().rowsMoved.connect(self.syncClipList)

        self.listHeaderButtonL = QPushButton(self)
        self.listHeaderButtonL.setObjectName('listheaderbutton-left')
        self.listHeaderButtonL.setFlat(True)
        self.listHeaderButtonL.clicked.connect(self.setClipIndexLayout)
        self.listHeaderButtonL.setCursor(Qt.PointingHandCursor)
        self.listHeaderButtonL.setFixedSize(14, 14)
        self.listHeaderButtonL.setToolTip('Move to left')
        self.listHeaderButtonL.setStatusTip('Move the Clip Index list to the left side of player')
        self.listHeaderButtonR = QPushButton(self)
        self.listHeaderButtonR.setObjectName('listheaderbutton-right')
        self.listHeaderButtonR.setFlat(True)
        self.listHeaderButtonR.clicked.connect(self.setClipIndexLayout)
        self.listHeaderButtonR.setCursor(Qt.PointingHandCursor)
        self.listHeaderButtonR.setFixedSize(14, 14)
        self.listHeaderButtonR.setToolTip('Move to right')
        self.listHeaderButtonR.setStatusTip('Move the Clip Index list to the right side of player')
        listheaderLayout = QHBoxLayout()
        listheaderLayout.setContentsMargins(6, 5, 6, 5)
        listheaderLayout.addWidget(self.listHeaderButtonL)
        listheaderLayout.addStretch(1)
        listheaderLayout.addWidget(self.listHeaderButtonR)
        self.listheader = QWidget(self)
        self.listheader.setObjectName('listheader')
        self.listheader.setFixedWidth(self.cliplist.width())
        self.listheader.setLayout(listheaderLayout)
        self._initClipIndexHeader()

        self.runtimeLabel = QLabel('<div align="right">00:00:00</div>', self)
        self.runtimeLabel.setObjectName('runtimeLabel')
        self.runtimeLabel.setToolTip('total runtime: 00:00:00')
        self.runtimeLabel.setStatusTip('total running time: 00:00:00')

        self.clipindex_add = QPushButton(self)
        self.clipindex_add.setObjectName('clipadd')
        self.clipindex_add.clicked.connect(self.addExternalClips)
        self.clipindex_add.setToolTip('Add clips')
        self.clipindex_add.setStatusTip('Add one or more files to an existing project or an empty list if you are only '
                                        'joining files')
        self.clipindex_add.setCursor(Qt.PointingHandCursor)
        self.clipindex_remove = QPushButton(self)
        self.clipindex_remove.setObjectName('clipremove')
        self.clipindex_remove.setToolTip('Remove clips')
        self.clipindex_remove.setStatusTip('Remove clips from your index')
        self.clipindex_remove.setLayoutDirection(Qt.RightToLeft)
        self.clipindex_remove.setMenu(self.clipindex_removemenu)
        self.clipindex_remove.setCursor(Qt.PointingHandCursor)
        if sys.platform in {'win32', 'darwin'}:
            self.clipindex_add.setStyle(QStyleFactory.create('Fusion'))
            self.clipindex_remove.setStyle(QStyleFactory.create('Fusion'))

        clipindex_layout = QHBoxLayout()
        clipindex_layout.setSpacing(1)
        clipindex_layout.setContentsMargins(0, 0, 0, 0)
        clipindex_layout.addWidget(self.clipindex_add)
        clipindex_layout.addSpacing(1)
        clipindex_layout.addWidget(self.clipindex_remove)
        clipindexTools = QWidget(self)
        clipindexTools.setObjectName('clipindextools')
        clipindexTools.setLayout(clipindex_layout)

        self.clipindexLayout = QVBoxLayout()
        self.clipindexLayout.setSpacing(0)
        self.clipindexLayout.setContentsMargins(0, 0, 0, 0)
        self.clipindexLayout.addWidget(self.listheader)
        self.clipindexLayout.addWidget(self.cliplist)
        self.clipindexLayout.addWidget(self.runtimeLabel)
        self.clipindexLayout.addSpacing(3)
        self.clipindexLayout.addWidget(clipindexTools)

        self.videoLayout = QHBoxLayout()
        self.videoLayout.setContentsMargins(0, 0, 0, 0)
        if self.indexLayout == 'left':
            self.videoLayout.addLayout(self.clipindexLayout)
            self.videoLayout.addSpacing(10)
            self.videoLayout.addWidget(self.novideoWidget)
        else:
            self.videoLayout.addWidget(self.novideoWidget)
            self.videoLayout.addSpacing(10)
            self.videoLayout.addLayout(self.clipindexLayout)

        self.timeCounter = VCTimeCounter(self)
        self.timeCounter.timeChanged.connect(lambda newtime: self.setPosition(newtime.msecsSinceStartOfDay()))
        self.frameCounter = VCFrameCounter(self)
        self.frameCounter.setReadOnly(True)

        countersLayout = QHBoxLayout()
        countersLayout.setContentsMargins(0, 0, 0, 0)
        countersLayout.addStretch(1)
        # noinspection PyArgumentList
        countersLayout.addWidget(QLabel('TIME:', objectName='tcLabel'))
        countersLayout.addWidget(self.timeCounter)
        countersLayout.addStretch(1)
        # noinspection PyArgumentList
        countersLayout.addWidget(QLabel('FRAME:', objectName='fcLabel'))
        countersLayout.addWidget(self.frameCounter)
        countersLayout.addStretch(1)

        countersWidget = QWidget(self)
        countersWidget.setObjectName('counterwidgets')
        countersWidget.setContentsMargins(0, 0, 0, 0)
        countersWidget.setLayout(countersLayout)
        countersWidget.setMaximumHeight(28)

        self.mpvWidget = self.getMPV(self)

        self.videoplayerLayout = QVBoxLayout()
        self.videoplayerLayout.setSpacing(0)
        self.videoplayerLayout.setContentsMargins(0, 0, 0, 0)
        self.videoplayerLayout.addWidget(self.mpvWidget)
        self.videoplayerLayout.addWidget(countersWidget)

        self.videoplayerWidget = QFrame(self)
        self.videoplayerWidget.setObjectName('videoplayer')
        self.videoplayerWidget.setFrameStyle(QFrame.Box | QFrame.Sunken)
        self.videoplayerWidget.setLineWidth(0)
        self.videoplayerWidget.setMidLineWidth(0)
        self.videoplayerWidget.setVisible(False)
        self.videoplayerWidget.setLayout(self.videoplayerLayout)

        # noinspection PyArgumentList
        self.thumbnailsButton = QPushButton(self, flat=True, checkable=True, objectName='thumbnailsButton',
                                            statusTip='Toggle timeline thumbnails', cursor=Qt.PointingHandCursor,
                                            toolTip='Toggle thumbnails')
        self.thumbnailsButton.setFixedSize(32, 29 if self.theme == 'dark' else 31)
        self.thumbnailsButton.setChecked(self.timelineThumbs)
        self.thumbnailsButton.toggled.connect(self.toggleThumbs)
        if self.timelineThumbs:
            self.seekSlider.setObjectName('nothumbs')

        # noinspection PyArgumentList
        self.osdButton = QPushButton(self, flat=True, checkable=True, objectName='osdButton', toolTip='Toggle OSD',
                                     statusTip='Toggle on-screen display', cursor=Qt.PointingHandCursor)
        self.osdButton.setFixedSize(31, 29 if self.theme == 'dark' else 31)
        self.osdButton.setChecked(self.enableOSD)
        self.osdButton.toggled.connect(self.toggleOSD)

        # noinspection PyArgumentList
        self.consoleButton = QPushButton(self, flat=True, checkable=True, objectName='consoleButton',
                                         statusTip='Toggle console window', toolTip='Toggle console',
                                         cursor=Qt.PointingHandCursor)
        self.consoleButton.setFixedSize(31, 29 if self.theme == 'dark' else 31)
        self.consoleButton.setChecked(self.showConsole)
        self.consoleButton.toggled.connect(self.toggleConsole)
        if self.showConsole:
            self.mpvWidget.setLogLevel('v')
            os.environ['DEBUG'] = '1'
            self.parent.console.show()

        # noinspection PyArgumentList
        self.chaptersButton = QPushButton(self, flat=True, checkable=True, objectName='chaptersButton',
                                          statusTip='Automatically create chapters per clip', toolTip='Create chapters',
                                          cursor=Qt.PointingHandCursor)
        self.chaptersButton.setFixedSize(31, 29 if self.theme == 'dark' else 31)
        self.chaptersButton.setChecked(self.createChapters)
        self.chaptersButton.toggled.connect(self.toggleChapters)

        # noinspection PyArgumentList
        self.smartcutButton = QPushButton(self, flat=True, checkable=True, objectName='smartcutButton',
                                          toolTip='Toggle SmartCut', statusTip='Toggle frame accurate cutting',
                                          cursor=Qt.PointingHandCursor)
        self.smartcutButton.setFixedSize(32, 29 if self.theme == 'dark' else 31)
        self.smartcutButton.setChecked(self.smartcut)
        self.smartcutButton.toggled.connect(self.toggleSmartCut)

        # noinspection PyArgumentList
        self.muteButton = QPushButton(objectName='muteButton', icon=self.unmuteIcon, flat=True, toolTip='Mute',
                                      statusTip='Toggle audio mute', iconSize=QSize(16, 16), clicked=self.muteAudio,
                                      cursor=Qt.PointingHandCursor)

        # noinspection PyArgumentList
        self.volSlider = VCVolumeSlider(orientation=Qt.Horizontal, toolTip='Volume', statusTip='Adjust volume level',
                                        cursor=Qt.PointingHandCursor, value=self.parent.startupvol, minimum=0,
                                        maximum=130, minimumHeight=22, sliderMoved=self.setVolume)

        # noinspection PyArgumentList
        self.fullscreenButton = QPushButton(objectName='fullscreenButton', icon=self.fullscreenIcon, flat=True,
                                            toolTip='Toggle fullscreen', statusTip='Switch to fullscreen video',
                                            iconSize=QSize(14, 14), clicked=self.toggleFullscreen,
                                            cursor=Qt.PointingHandCursor, enabled=False)

        # noinspection PyArgumentList
        self.settingsButton = QPushButton(self, toolTip='Settings', cursor=Qt.PointingHandCursor, flat=True,
                                          statusTip='Configure application settings',
                                          objectName='settingsButton', clicked=self.showSettings)
        self.settingsButton.setFixedSize(QSize(33, 32))

        # noinspection PyArgumentList
        self.streamsButton = QPushButton(self, toolTip='Media streams', cursor=Qt.PointingHandCursor, flat=True,
                                         statusTip='Select the media streams to be included',
                                         objectName='streamsButton', clicked=self.selectStreams,
                                         enabled=False)
        self.streamsButton.setFixedSize(QSize(33, 32))

        # noinspection PyArgumentList
        self.mediainfoButton = QPushButton(self, toolTip='Media information', cursor=Qt.PointingHandCursor, flat=True,
                                           statusTip='View technical details about current media',
                                           objectName='mediainfoButton', clicked=self.mediaInfo, enabled=False)
        self.mediainfoButton.setFixedSize(QSize(33, 32))

        # noinspection PyArgumentList
        self.menuButton = QPushButton(self, toolTip='Menu', cursor=Qt.PointingHandCursor, flat=True,
                                      objectName='menuButton', clicked=self.showAppMenu, statusTip='View menu options')
        self.menuButton.setFixedSize(QSize(33, 32))

        audioLayout = QHBoxLayout()
        audioLayout.setContentsMargins(0, 0, 0, 0)
        audioLayout.addWidget(self.muteButton)
        audioLayout.addSpacing(5)
        audioLayout.addWidget(self.volSlider)
        audioLayout.addSpacing(5)
        audioLayout.addWidget(self.fullscreenButton)

        self.toolbar_open = VCToolBarButton('Open Media', 'Open and load a media file to begin', parent=self)
        self.toolbar_open.clicked.connect(self.openMedia)
        self.toolbar_play = VCToolBarButton('Play Media', 'Play currently loaded media file', parent=self)
        self.toolbar_play.setEnabled(False)
        self.toolbar_play.clicked.connect(self.playMedia)
        self.toolbar_start = VCToolBarButton('Start Clip', 'Start a new clip from the current timeline position',
                                             parent=self)
        self.toolbar_start.setEnabled(False)
        self.toolbar_start.clicked.connect(self.clipStart)
        self.toolbar_end = VCToolBarButton('End Clip', 'End a new clip at the current timeline position', parent=self)
        self.toolbar_end.setEnabled(False)
        self.toolbar_end.clicked.connect(self.clipEnd)
        self.toolbar_save = VCToolBarButton('Save Media', 'Save clips to a new media file', parent=self)
        self.toolbar_save.setObjectName('savebutton')
        self.toolbar_save.setEnabled(False)
        self.toolbar_save.clicked.connect(self.saveMedia)

        toolbarLayout = QHBoxLayout()
        toolbarLayout.setContentsMargins(0, 0, 0, 0)
        toolbarLayout.addStretch(1)
        toolbarLayout.addWidget(self.toolbar_open)
        toolbarLayout.addStretch(1)
        toolbarLayout.addWidget(self.toolbar_play)
        toolbarLayout.addStretch(1)
        toolbarLayout.addWidget(self.toolbar_start)
        toolbarLayout.addStretch(1)
        toolbarLayout.addWidget(self.toolbar_end)
        toolbarLayout.addStretch(1)
        toolbarLayout.addWidget(self.toolbar_save)
        toolbarLayout.addStretch(1)

        self.toolbarGroup = QGroupBox()
        self.toolbarGroup.setLayout(toolbarLayout)
        self.toolbarGroup.setStyleSheet('QGroupBox { border: 0; }')

        self.setToolBarStyle(self.settings.value('toolbarLabels', 'beside', type=str))

        togglesLayout = QHBoxLayout()
        togglesLayout.setSpacing(0)
        togglesLayout.setContentsMargins(0, 0, 0, 0)
        togglesLayout.addWidget(self.consoleButton)
        togglesLayout.addWidget(self.osdButton)
        togglesLayout.addWidget(self.thumbnailsButton)
        togglesLayout.addWidget(self.chaptersButton)
        togglesLayout.addWidget(self.smartcutButton)
        togglesLayout.addStretch(1)

        settingsLayout = QHBoxLayout()
        settingsLayout.setSpacing(0)
        settingsLayout.setContentsMargins(0, 0, 0, 0)
        settingsLayout.addWidget(self.settingsButton)
        settingsLayout.addSpacing(5)
        settingsLayout.addWidget(self.streamsButton)
        settingsLayout.addSpacing(5)
        settingsLayout.addWidget(self.mediainfoButton)
        settingsLayout.addSpacing(5)
        settingsLayout.addWidget(self.menuButton)

        groupLayout = QVBoxLayout()
        groupLayout.addLayout(audioLayout)
        groupLayout.addSpacing(10)
        groupLayout.addLayout(settingsLayout)

        controlsLayout = QHBoxLayout()
        if sys.platform != 'darwin':
            controlsLayout.setContentsMargins(0, 0, 0, 0)
            controlsLayout.addSpacing(5)
        else:
            controlsLayout.setContentsMargins(10, 10, 10, 0)
        controlsLayout.addLayout(togglesLayout)
        controlsLayout.addSpacing(20)
        controlsLayout.addStretch(1)
        controlsLayout.addWidget(self.toolbarGroup)
        controlsLayout.addStretch(1)
        controlsLayout.addSpacing(20)
        controlsLayout.addLayout(groupLayout)
        if sys.platform != 'darwin':
            controlsLayout.addSpacing(5)

        layout = QVBoxLayout()  
        layout.setSpacing(0)
        layout.setContentsMargins(10, 10, 10, 0)
        layout.addLayout(self.videoLayout)
        layout.addWidget(self.sliderWidget)
        layout.addSpacing(5)
        layout.addLayout(controlsLayout)

        self.setLayout(layout)
        self.seekSlider.initStyle()

    @pyqtSlot()
    def showAppMenu(self) -> None:
        pos = self.menuButton.mapToGlobal(self.menuButton.rect().topLeft())
        pos.setX(pos.x() - self.appmenu.sizeHint().width() + 30)
        pos.setY(pos.y() - 28)
        self.appmenu.popup(pos, self.quitAction)

    def initTheme(self) -> None:
        qApp.setStyle(VideoStyleDark() if self.theme == 'dark' else VideoStyleLight())
        self.fonts = [
            QFontDatabase.addApplicationFont(':/fonts/FuturaLT.ttf'),
            QFontDatabase.addApplicationFont(':/fonts/NotoSans-Bold.ttf'),
            QFontDatabase.addApplicationFont(':/fonts/NotoSans-Regular.ttf')
        ]
        self.style().loadQSS(self.theme)
        QApplication.setFont(QFont('Noto Sans', 12 if sys.platform == 'darwin' else 10, 300))

    def getMPV(self, parent: QWidget=None, file: str=None, start: float=0, pause: bool=True, mute: bool=False,
               volume: int=None) -> mpvWidget:
        widget = mpvWidget(
            parent=parent,
            file=file,
            vo='opengl-cb',
            pause=pause,
            start=start,
            mute=mute,
            keep_open='always',
            idle=True,
            osd_font=self._osdfont,
            osd_level=0,
            osd_align_x='left',
            osd_align_y='top',
            cursor_autohide=False,
            input_cursor=False,
            input_default_bindings=False,
            stop_playback_on_init_failure=False,
            input_vo_keyboard=False,
            sub_auto=False,
            sid=False,
            video_sync='display-vdrop',
            audio_file_auto=False,
            quiet=True,
            volume=volume if volume is not None else self.parent.startupvol,
            opengl_pbo=self.enablePBO,
            keepaspect=self.keepRatio,
            hwdec=('auto' if self.hardwareDecoding else 'no'))
        widget.durationChanged.connect(self.on_durationChanged)
        widget.positionChanged.connect(self.on_positionChanged)
        return widget

    def _initNoVideo(self) -> None:
        self.novideoWidget = QWidget(self)
        self.novideoWidget.setObjectName('novideoWidget')
        self.novideoWidget.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        openmediaLabel = VCBlinkText('open media to begin', self)
        openmediaLabel.setAlignment(Qt.AlignHCenter)
        _version = 'v{}'.format(qApp.applicationVersion())
        if self.parent.flatpak:
            _version += ' <font size="-1">- FLATPAK</font>'
        versionLabel = QLabel(_version, self)
        versionLabel.setObjectName('novideoversion')
        versionLabel.setAlignment(Qt.AlignRight)
        versionLayout = QHBoxLayout()
        versionLayout.setSpacing(0)
        versionLayout.setContentsMargins(0, 0, 10, 8)
        versionLayout.addWidget(versionLabel)
        novideoLayout = QVBoxLayout(self.novideoWidget)
        novideoLayout.setSpacing(0)
        novideoLayout.setContentsMargins(0, 0, 0, 0)
        novideoLayout.addStretch(20)
        novideoLayout.addWidget(openmediaLabel)
        novideoLayout.addStretch(1)
        novideoLayout.addLayout(versionLayout)

    def _initIcons(self) -> None:
        self.appIcon = qApp.windowIcon()
        self.muteIcon = QIcon(':/images/{}/muted.png'.format(self.theme))
        self.unmuteIcon = QIcon(':/images/{}/unmuted.png'.format(self.theme))
        self.chapterIcon = QIcon(':/images/chapters.png')
        self.upIcon = QIcon(':/images/up.png')
        self.downIcon = QIcon(':/images/down.png')
        self.removeIcon = QIcon(':/images/remove.png')
        self.removeAllIcon = QIcon(':/images/remove-all.png')
        self.openProjectIcon = QIcon(':/images/open.png')
        self.saveProjectIcon = QIcon(':/images/save.png')
        self.filtersIcon = QIcon(':/images/filters.png')
        self.mediaInfoIcon = QIcon(':/images/info.png')
        self.streamsIcon = QIcon(':/images/streams.png')
        self.changelogIcon = QIcon(':/images/changelog.png')
        self.viewLogsIcon = QIcon(':/images/viewlogs.png')
        self.updateCheckIcon = QIcon(':/images/update.png')
        self.keyRefIcon = QIcon(':/images/keymap.png')
        self.fullscreenIcon = QIcon(':/images/{}/fullscreen.png'.format(self.theme))
        self.settingsIcon = QIcon(':/images/settings.png')
        self.quitIcon = QIcon(':/images/quit.png')

    # noinspection PyArgumentList
    def _initActions(self) -> None:
        self.moveItemUpAction = QAction(self.upIcon, 'Move clip up', self, statusTip='Move clip position up in list',
                                        triggered=self.moveItemUp, enabled=False)
        self.moveItemDownAction = QAction(self.downIcon, 'Move clip down', self, triggered=self.moveItemDown,
                                          statusTip='Move clip position down in list', enabled=False)
        self.removeItemAction = QAction(self.removeIcon, 'Remove selected clip', self, triggered=self.removeItem,
                                        statusTip='Remove selected clip from list', enabled=False)
        self.removeAllAction = QAction(self.removeAllIcon, 'Remove all clips', self, triggered=self.clearList,
                                       statusTip='Remove all clips from list', enabled=False)
        self.editChapterAction = QAction(self.chapterIcon, 'Edit chapter name', self, triggered=self.editChapter,
                                         statusTip='Edit the selected chapter name', enabled=False)
        self.streamsAction = QAction(self.streamsIcon, 'Media streams', self, triggered=self.selectStreams,
                                     statusTip='Select the media streams to be included', enabled=False)
        self.mediainfoAction = QAction(self.mediaInfoIcon, 'Media information', self, triggered=self.mediaInfo,
                                       statusTip='View technical details about current media', enabled=False)
        self.openProjectAction = QAction(self.openProjectIcon, 'Open project file', self, triggered=self.openProject,
                                         statusTip='Open a previously saved project file (*.vcp or *.edl)',
                                         enabled=True)
        self.saveProjectAction = QAction(self.saveProjectIcon, 'Save project file', self, triggered=self.saveProject,
                                         statusTip='Save current work to a project file (*.vcp or *.edl)',
                                         enabled=False)
        self.changelogAction = QAction(self.changelogIcon, 'View changelog', self, triggered=self.viewChangelog,
                                       statusTip='View log of changes')
        self.viewLogsAction = QAction(self.viewLogsIcon, 'View log file', self, triggered=VideoCutter.viewLogs,
                                      statusTip='View the application\'s log file')
        self.updateCheckAction = QAction(self.updateCheckIcon, 'Check for updates...', self,
                                         statusTip='Check for application updates', triggered=self.updater.check)
        self.aboutQtAction = QAction('About Qt', self, triggered=qApp.aboutQt, statusTip='About Qt')
        self.aboutAction = QAction('About {}'.format(qApp.applicationName()), self, triggered=self.aboutApp,
                                   statusTip='About {}'.format(qApp.applicationName()))
        self.keyRefAction = QAction(self.keyRefIcon, 'Keyboard shortcuts', self, triggered=self.showKeyRef,
                                    statusTip='View shortcut key bindings')
        self.settingsAction = QAction(self.settingsIcon, 'Settings', self, triggered=self.showSettings,
                                      statusTip='Configure application settings')
        self.fullscreenAction = QAction(self.fullscreenIcon, 'Toggle fullscreen', self, triggered=self.toggleFullscreen,
                                        statusTip='Toggle fullscreen display mode', enabled=False)
        self.quitAction = QAction(self.quitIcon, 'Quit', self, triggered=self.parent.close,
                                  statusTip='Quit the application')

    @property
    def _filtersMenu(self) -> QMenu:
        menu = QMenu('Video filters', self)
        self.blackdetectAction = VCFilterMenuAction(QPixmap(':/images/blackdetect.png'), 'BLACKDETECT',
                                                    'Create clips via black frame detection',
                                                    'Useful for skipping commercials or detecting scene transitions',
                                                    self)
        if sys.platform == 'darwin':
            self.blackdetectAction.triggered.connect(lambda: self.configFilters(VideoFilter.BLACKDETECT),
                                                     Qt.QueuedConnection)
        else:
            self.blackdetectAction.triggered.connect(lambda: self.configFilters(VideoFilter.BLACKDETECT),
                                                     Qt.DirectConnection)
        self.blackdetectAction.setEnabled(False)
        menu.setIcon(self.filtersIcon)
        menu.addAction(self.blackdetectAction)
        return menu

    def _initMenus(self) -> None:
        self.appmenu.addAction(self.openProjectAction)
        self.appmenu.addAction(self.saveProjectAction)
        self.appmenu.addSeparator()
        self.appmenu.addMenu(self._filtersMenu)
        self.appmenu.addSeparator()
        self.appmenu.addAction(self.fullscreenAction)
        self.appmenu.addAction(self.streamsAction)
        self.appmenu.addAction(self.mediainfoAction)
        self.appmenu.addAction(self.keyRefAction)
        self.appmenu.addSeparator()
        self.appmenu.addAction(self.settingsAction)
        self.appmenu.addSeparator()
        self.appmenu.addAction(self.viewLogsAction)
        self.appmenu.addAction(self.updateCheckAction)
        self.appmenu.addSeparator()
        self.appmenu.addAction(self.changelogAction)
        self.appmenu.addAction(self.aboutQtAction)
        self.appmenu.addAction(self.aboutAction)
        self.appmenu.addSeparator()
        self.appmenu.addAction(self.quitAction)

        self.clipindex_contextmenu.addAction(self.editChapterAction)
        self.clipindex_contextmenu.addSeparator()
        self.clipindex_contextmenu.addAction(self.moveItemUpAction)
        self.clipindex_contextmenu.addAction(self.moveItemDownAction)
        self.clipindex_contextmenu.addSeparator()
        self.clipindex_contextmenu.addAction(self.removeItemAction)
        self.clipindex_contextmenu.addAction(self.removeAllAction)

        self.clipindex_removemenu.addActions([self.removeItemAction, self.removeAllAction])
        self.clipindex_removemenu.aboutToShow.connect(self.initRemoveMenu)

        if sys.platform in {'win32', 'darwin'}:
            self.appmenu.setStyle(QStyleFactory.create('Fusion'))
            self.clipindex_contextmenu.setStyle(QStyleFactory.create('Fusion'))
            self.clipindex_removemenu.setStyle(QStyleFactory.create('Fusion'))

    def _initClipIndexHeader(self) -> None:
        if self.indexLayout == 'left':
            self.listHeaderButtonL.setVisible(False)
            self.listHeaderButtonR.setVisible(True)
        else:
            self.listHeaderButtonL.setVisible(True)
            self.listHeaderButtonR.setVisible(False)

    @pyqtSlot()
    def setClipIndexLayout(self) -> None:
        self.indexLayout = 'left' if self.indexLayout == 'right' else 'right'
        self.settings.setValue('indexLayout', self.indexLayout)
        left = self.videoLayout.takeAt(0)
        spacer = self.videoLayout.takeAt(0)
        right = self.videoLayout.takeAt(0)
        if isinstance(left, QVBoxLayout):
            if self.indexLayout == 'left':
                self.videoLayout.addItem(left)
                self.videoLayout.addItem(spacer)
                self.videoLayout.addItem(right)
            else:
                self.videoLayout.addItem(right)
                self.videoLayout.addItem(spacer)
                self.videoLayout.addItem(left)
        else:
            if self.indexLayout == 'left':
                self.videoLayout.addItem(right)
                self.videoLayout.addItem(spacer)
                self.videoLayout.addItem(left)
            else:
                self.videoLayout.addItem(left)
                self.videoLayout.addItem(spacer)
                self.videoLayout.addItem(right)
        self._initClipIndexHeader()

    def setToolBarStyle(self, labelstyle: str = 'beside') -> None:
        buttonlist = self.toolbarGroup.findChildren(VCToolBarButton)
        [button.setLabelStyle(labelstyle) for button in buttonlist]

    def setRunningTime(self, runtime: str) -> None:
        self.runtimeLabel.setText('<div align="right">{}</div>'.format(runtime))
        self.runtimeLabel.setToolTip('total runtime: {}'.format(runtime))
        self.runtimeLabel.setStatusTip('total running time: {}'.format(runtime))

    def getFileDialogOptions(self) -> QFileDialog.Options:
        options = QFileDialog.HideNameFilterDetails
        if not self.nativeDialogs:
            options |= QFileDialog.DontUseNativeDialog
        # noinspection PyTypeChecker
        return options

    @pyqtSlot()
    def showSettings(self):
        settingsDialog = SettingsDialog(self.videoService, self)
        settingsDialog.exec_()

    @pyqtSlot()
    def initRemoveMenu(self):
        self.removeItemAction.setEnabled(False)
        self.removeAllAction.setEnabled(False)
        if self.cliplist.count():
            self.removeAllAction.setEnabled(True)
            if len(self.cliplist.selectedItems()):
                self.removeItemAction.setEnabled(True)

    def itemMenu(self, pos: QPoint) -> None:
        globalPos = self.cliplist.mapToGlobal(pos)
        self.editChapterAction.setEnabled(False)
        self.moveItemUpAction.setEnabled(False)
        self.moveItemDownAction.setEnabled(False)
        self.initRemoveMenu()
        index = self.cliplist.currentRow()
        if index != -1:
            if len(self.cliplist.selectedItems()):
                self.editChapterAction.setEnabled(self.createChapters)
            if not self.inCut:
                if index > 0:
                    self.moveItemUpAction.setEnabled(True)
                if index < self.cliplist.count() - 1:
                    self.moveItemDownAction.setEnabled(True)
        self.clipindex_contextmenu.exec_(globalPos)

    def editChapter(self) -> None:
        index = self.cliplist.currentRow()
        name = self.clipTimes[index][4]
        name = name if name is not None else 'Chapter {}'.format(index + 1)
        dialog = VCInputDialog(self, 'Edit chapter name', 'Chapter name:', name)
        dialog.accepted.connect(lambda: self.on_editChapter(index, dialog.input.text()))
        dialog.exec_()

    def on_editChapter(self, index: int, text: str) -> None:
        self.clipTimes[index][4] = text
        self.renderClipIndex()

    def moveItemUp(self) -> None:
        index = self.cliplist.currentRow()
        if index != -1:
            tmpItem = self.clipTimes[index]
            del self.clipTimes[index]
            self.clipTimes.insert(index - 1, tmpItem)
            self.showText('clip moved up')
            self.renderClipIndex()

    def moveItemDown(self) -> None:
        index = self.cliplist.currentRow()
        if index != -1:
            tmpItem = self.clipTimes[index]
            del self.clipTimes[index]
            self.clipTimes.insert(index + 1, tmpItem)
            self.showText('clip moved down')
            self.renderClipIndex()

    def removeItem(self) -> None:
        index = self.cliplist.currentRow()
        if self.mediaAvailable:
            if self.inCut and index == self.cliplist.count() - 1:
                self.inCut = False
                self.initMediaControls()
        elif len(self.clipTimes) == 0:
            self.initMediaControls(False)
        del self.clipTimes[index]
        self.cliplist.takeItem(index)
        self.showText('clip removed')
        self.renderClipIndex()

    def clearList(self) -> None:
        self.clipTimes.clear()
        self.cliplist.clear()
        self.showText('all clips cleared')
        if self.mediaAvailable:
            self.inCut = False
            self.initMediaControls(True)
        else:
            self.initMediaControls(False)
        self.renderClipIndex()

    def projectFilters(self, savedialog: bool = False) -> str:
        if savedialog:
            return 'VidCutter Project (*.vcp);;MPlayer EDL (*.edl)'
        elif self.mediaAvailable:
            return 'Project files (*.edl *.vcp);;VidCutter Project (*.vcp);;MPlayer EDL (*.edl);;All files (*)'
        else:
            return 'VidCutter Project (*.vcp);;All files (*)'

    @staticmethod
    def mediaFilters(initial: bool = False) -> str:
        filters = 'All media files (*.{})'.format(' *.'.join(VideoService.config.filters.get('all')))
        if initial:
            return filters
        filters += ';;{};;All files (*)'.format(';;'.join(VideoService.config.filters.get('types')))
        return filters

    def openMedia(self) -> Optional[Callable]:
        cancel, callback = self.saveWarning()
        if cancel:
            if callback is not None:
                return callback()
            else:
                return
        filename, _ = QFileDialog.getOpenFileName(
            parent=self.parent,
            caption='Open media file',
            filter=self.mediaFilters(),
            initialFilter=self.mediaFilters(True),
            directory=(self.lastFolder if os.path.exists(self.lastFolder) else QDir.homePath()),
            options=self.getFileDialogOptions())
        if filename is not None and len(filename.strip()):
            self.lastFolder = QFileInfo(filename).absolutePath()
            self.loadMedia(filename)

    # noinspection PyUnusedLocal
    def openProject(self, checked: bool = False, project_file: str = None) -> Optional[Callable]:
        cancel, callback = self.saveWarning()
        if cancel:
            if callback is not None:
                return callback()
            else:
                return
        initialFilter = 'Project files (*.edl *.vcp)' if self.mediaAvailable else 'VidCutter Project (*.vcp)'
        if project_file is None:
            project_file, _ = QFileDialog.getOpenFileName(
                parent=self.parent,
                caption='Open project file',
                filter=self.projectFilters(),
                initialFilter=initialFilter,
                directory=(self.lastFolder if os.path.exists(self.lastFolder) else QDir.homePath()),
                options=self.getFileDialogOptions())
        if project_file is not None and len(project_file.strip()):
            if project_file != os.path.join(QDir.tempPath(), self.parent.TEMP_PROJECT_FILE):
                self.lastFolder = QFileInfo(project_file).absolutePath()
            file = QFile(project_file)
            info = QFileInfo(file)
            project_type = info.suffix()
            if not file.open(QFile.ReadOnly | QFile.Text):
                QMessageBox.critical(self.parent, 'Open project file',
                                     'Cannot read project file {0}:\n\n{1}'.format(project_file, file.errorString()))
                return
            qApp.setOverrideCursor(Qt.WaitCursor)
            self.clipTimes.clear()
            linenum = 1
            while not file.atEnd():
                # noinspection PyUnresolvedReferences
                line = file.readLine().trimmed()
                if line.length() > 0:
                    try:
                        line = line.data().decode()
                    except UnicodeDecodeError:
                        qApp.restoreOverrideCursor()
                        self.logger.error('Invalid project file was selected', exc_info=True)
                        sys.stderr.write('Invalid project file was selected')
                        QMessageBox.critical(self.parent, 'Invalid project file',
                                             'Could not make sense of the selected project file. Try viewing it in a '
                                             'text editor to ensure it is valid and not corrupted.')
                        return
                    if project_type == 'vcp' and linenum == 1:
                        self.loadMedia(line)
                        time.sleep(1)
                    else:
                        mo = self.project_files[project_type].match(line)
                        if mo:
                            start, stop, _, chapter = mo.groups()
                            clip_start = self.delta2QTime(float(start))
                            clip_end = self.delta2QTime(float(stop))
                            clip_image = self.captureImage(self.currentMedia, clip_start)
                            if project_type == 'vcp' and self.createChapters and len(chapter):
                                chapter = chapter[1:len(chapter) - 1]
                                if not len(chapter):
                                    chapter = None
                            else:
                                chapter = None
                            self.clipTimes.append([clip_start, clip_end, clip_image, '', chapter])
                        else:
                            qApp.restoreOverrideCursor()
                            QMessageBox.critical(self.parent, 'Invalid project file',
                                                 'Invalid entry at line {0}:\n\n{1}'.format(linenum, line))
                            return
                linenum += 1
            self.toolbar_start.setEnabled(True)
            self.toolbar_end.setDisabled(True)
            self.seekSlider.setRestrictValue(0, False)
            self.blackdetectAction.setEnabled(True)
            self.inCut = False
            self.newproject = True
            QTimer.singleShot(2000, self.selectClip)
            qApp.restoreOverrideCursor()
            if project_file != os.path.join(QDir.tempPath(), self.parent.TEMP_PROJECT_FILE):
                self.showText('project loaded')

    def saveProject(self, reboot: bool = False) -> None:
        if self.currentMedia is None:
            return
        if self.hasExternals():
            h2color = '#C681D5' if self.theme == 'dark' else '#642C68'
            acolor = '#EA95FF' if self.theme == 'dark' else '#441D4E'
            nosavetext = '''
                <style>
                    h2 {{
                        color: {h2color};
                        font-family: "Futura LT", sans-serif;
                        font-weight: normal;
                    }}
                    a {{
                        color: {acolor};
                        text-decoration: none;
                        font-weight: bold;
                    }}
                </style>
                <table border="0" cellpadding="6" cellspacing="0" width="350">
                    <tr>
                        <td><h2>Cannot save your current project</h2></td>
                    </tr>
                    <tr>
                        <td>
                            <p>Cannot save project containing external media files. Remove
                            all media files you have externally added and try again.</p>
                        </td>
                    </tr>
                </table>'''.format(**locals())
            nosave = QMessageBox(QMessageBox.Critical, 'Cannot save project', nosavetext, parent=self.parent)
            nosave.setStandardButtons(QMessageBox.Ok)
            nosave.exec_()
            return
        project_file, _ = os.path.splitext(self.currentMedia)
        if reboot:
            project_save = os.path.join(QDir.tempPath(), self.parent.TEMP_PROJECT_FILE)
            ptype = 'VidCutter Project (*.vcp)'
        else:
            project_save, ptype = QFileDialog.getSaveFileName(
                parent=self.parent,
                caption='Save project',
                directory='{}.vcp'.format(project_file),
                filter=self.projectFilters(True),
                initialFilter='VidCutter Project (*.vcp)',
                options=self.getFileDialogOptions())
        if project_save is not None and len(project_save.strip()):
            file = QFile(project_save)
            if not file.open(QFile.WriteOnly | QFile.Text):
                QMessageBox.critical(self.parent, 'Cannot save project',
                                     'Cannot save project file at {0}:\n\n{1}'.format(project_save, file.errorString()))
                return
            qApp.setOverrideCursor(Qt.WaitCursor)
            if ptype == 'VidCutter Project (*.vcp)':
                # noinspection PyUnresolvedReferences
                QTextStream(file) << '{}\n'.format(self.currentMedia)
            for clip in self.clipTimes:
                start_time = timedelta(hours=clip[0].hour(), minutes=clip[0].minute(), seconds=clip[0].second(),
                                       milliseconds=clip[0].msec())
                stop_time = timedelta(hours=clip[1].hour(), minutes=clip[1].minute(), seconds=clip[1].second(),
                                      milliseconds=clip[1].msec())
                if ptype == 'VidCutter Project (*.vcp)':
                    if self.createChapters:
                        chapter = '"{}"'.format(clip[4]) if clip[4] is not None else '""'
                    else:
                        chapter = ''
                    # noinspection PyUnresolvedReferences
                    QTextStream(file) << '{0}\t{1}\t{2}\t{3}\n'.format(self.delta2String(start_time),
                                                                       self.delta2String(stop_time), 0, chapter)
                else:
                    # noinspection PyUnresolvedReferences
                    QTextStream(file) << '{0}\t{1}\t{2}\n'.format(self.delta2String(start_time),
                                                                  self.delta2String(stop_time), 0)
            qApp.restoreOverrideCursor()
            self.projectSaved = True
            if not reboot:
                self.showText('project file saved')

    def loadMedia(self, filename: str) -> None:
        if not os.path.isfile(filename):
            return
        self.currentMedia = filename
        self.initMediaControls(True)
        self.projectDirty, self.projectSaved = False, False
        self.cliplist.clear()
        self.clipTimes.clear()
        self.totalRuntime = 0
        self.setRunningTime(self.delta2QTime(self.totalRuntime).toString(self.runtimeformat))
        self.seekSlider.clearRegions()
        self.taskbar.init()
        self.parent.setWindowTitle('{0} - {1}'.format(qApp.applicationName(), os.path.basename(self.currentMedia)))
        if not self.mediaAvailable:
            self.videoLayout.replaceWidget(self.novideoWidget, self.videoplayerWidget)
            self.novideoWidget.hide()
            self.novideoWidget.deleteLater()
            self.videoplayerWidget.show()
            self.mediaAvailable = True
        try:
            self.videoService.setMedia(self.currentMedia)
            self.seekSlider.setFocus()
            self.mpvWidget.play(self.currentMedia)
        except InvalidMediaException:
            qApp.restoreOverrideCursor()
            self.initMediaControls(False)
            self.logger.error('Could not load media file', exc_info=True)
            QMessageBox.critical(self.parent, 'Could not load media file',
                                 '<h3>Invalid media file selected</h3><p>All attempts to make sense of the file have '
                                 'failed. Try viewing it in another media player and if it plays as expected please '
                                 'report it as a bug. Use the link in the About VidCutter menu option for details '
                                 'and make sure to include your operating system, video card, the invalid media file '
                                 'and the version of VidCutter you are currently using.</p>')

    def setPlayButton(self, playing: bool=False) -> None:
        self.toolbar_play.setup('{} Media'.format('Pause' if playing else 'Play'),
                                'Pause currently playing media' if playing else 'Play currently loaded media',
                                True)

    def playMedia(self) -> None:
        playstate = self.mpvWidget.property('pause')
        self.setPlayButton(playstate)
        self.taskbar.setState(playstate)
        self.timeCounter.clearFocus()
        self.frameCounter.clearFocus()
        self.mpvWidget.pause()

    def showText(self, text: str, duration: int = 3, override: bool = False) -> None:
        if self.mediaAvailable:
            if not self.osdButton.isChecked() and not override:
                return
            if len(text.strip()):
                self.mpvWidget.showText(text, duration)

    def initMediaControls(self, flag: bool = True) -> None:
        self.toolbar_play.setEnabled(flag)
        self.toolbar_start.setEnabled(flag)
        self.toolbar_end.setEnabled(False)
        self.toolbar_save.setEnabled(False)
        self.streamsAction.setEnabled(flag)
        self.streamsButton.setEnabled(flag)
        self.mediainfoAction.setEnabled(flag)
        self.mediainfoButton.setEnabled(flag)
        self.fullscreenButton.setEnabled(flag)
        self.fullscreenAction.setEnabled(flag)
        self.seekSlider.clearRegions()
        self.blackdetectAction.setEnabled(flag)
        if flag:
            self.seekSlider.setRestrictValue(0)
        else:
            self.seekSlider.setValue(0)
            self.seekSlider.setRange(0, 0)
            self.timeCounter.reset()
            self.frameCounter.reset()
        self.openProjectAction.setEnabled(flag)
        self.saveProjectAction.setEnabled(False)

    @pyqtSlot(int)
    def setPosition(self, position: int) -> None:
        if position >= self.seekSlider.restrictValue:
            self.mpvWidget.seek(position / 1000)

    @pyqtSlot(float, int)
    def on_positionChanged(self, progress: float, frame: int) -> None:
        progress *= 1000
        if self.seekSlider.restrictValue < progress or progress == 0:
            self.seekSlider.setValue(int(progress))
            self.timeCounter.setTime(self.delta2QTime(round(progress)).toString(self.timeformat))
            self.frameCounter.setFrame(frame)
            if self.seekSlider.maximum() > 0:
                self.taskbar.setProgress(float(progress / self.seekSlider.maximum()), True)

    @pyqtSlot(float, int)
    def on_durationChanged(self, duration: float, frames: int) -> None:
        duration *= 1000
        self.seekSlider.setRange(0, int(duration))
        self.timeCounter.setDuration(self.delta2QTime(round(duration)).toString(self.timeformat))
        self.frameCounter.setFrameCount(frames)

    @pyqtSlot()
    @pyqtSlot(QListWidgetItem)
    def selectClip(self, item: QListWidgetItem = None) -> None:
        # noinspection PyBroadException
        try:
            row = self.cliplist.row(item) if item is not None else 0
            if item is None:
                self.cliplist.item(row).setSelected(True)
            if not len(self.clipTimes[row][3]):
                self.seekSlider.selectRegion(row)
                self.setPosition(self.clipTimes[row][0].msecsSinceStartOfDay())
        except Exception:
            self.doPass()

    def muteAudio(self) -> None:
        if self.mpvWidget.property('mute'):
            self.showText('audio enabled')
            self.muteButton.setIcon(self.unmuteIcon)
            self.muteButton.setToolTip('Mute')
        else:
            self.showText('audio disabled')
            self.muteButton.setIcon(self.muteIcon)
            self.muteButton.setToolTip('Unmute')
        self.mpvWidget.mute()

    def setVolume(self, vol: int) -> None:
        self.settings.setValue('volume', vol)
        if self.mediaAvailable:
            self.mpvWidget.volume(vol)

    @pyqtSlot(bool)
    def toggleThumbs(self, checked: bool) -> None:
        self.seekSlider.showThumbs = checked
        self.saveSetting('timelineThumbs', checked)
        if checked:
            self.showText('thumbnails enabled')
            self.seekSlider.initStyle()
            if self.mediaAvailable:
                self.seekSlider.reloadThumbs()
        else:
            self.showText('thumbnails disabled')
            self.seekSlider.removeThumbs()
            self.seekSlider.initStyle()

    @pyqtSlot(bool)
    def toggleConsole(self, checked: bool) -> None:
        if not hasattr(self, 'debugonstart'):
            self.debugonstart = os.getenv('DEBUG', False)
        if checked:
            self.mpvWidget.setLogLevel('v')
            os.environ['DEBUG'] = '1'
            self.parent.console.show()
        else:
            if not self.debugonstart:
                os.environ['DEBUG'] = '0'
                self.mpvWidget.setLogLevel('error')
            self.parent.console.hide()
        self.saveSetting('showConsole', checked)

    @pyqtSlot(bool)
    def toggleChapters(self, checked: bool) -> None:
        self.createChapters = checked
        self.saveSetting('chapters', self.createChapters)
        self.chaptersButton.setChecked(self.createChapters)
        self.showText('chapters {}'.format('enabled' if checked else 'disabled'))
        if checked:
            exist = False
            for clip in self.clipTimes:
                if clip[4] is not None:
                    exist = True
                    break
            if exist:
                chapterswarn = VCMessageBox('Restore chapter names', 'Chapter names found in memory',
                                            'Would you like to restore previously set chapter names?',
                                            buttons=QMessageBox.Yes | QMessageBox.No, parent=self)
                if chapterswarn.exec_() == QMessageBox.No:
                    for clip in self.clipTimes:
                        clip[4] = None
        self.renderClipIndex()

    @pyqtSlot(bool)
    def toggleSmartCut(self, checked: bool) -> None:
        self.smartcut = checked
        self.saveSetting('smartcut', self.smartcut)
        self.smartcutButton.setChecked(self.smartcut)
        self.showText('SmartCut {}'.format('enabled' if checked else 'disabled'))

    @pyqtSlot(list)
    def addScenes(self, scenes: List[list]) -> None:
        if len(scenes):
            [
                self.clipTimes.append([scene[0], scene[1], self.captureImage(self.currentMedia, scene[0]), '', None])
                for scene in scenes if len(scene)
            ]
            self.renderClipIndex()
        self.filterProgressBar.done(VCProgressDialog.Accepted)

    @pyqtSlot(VideoFilter)
    def configFilters(self, name: VideoFilter) -> None:
        if name == VideoFilter.BLACKDETECT:
            desc = '<p>Detect video intervals that are (almost) completely black. Can be useful to detect chapter ' \
                   'transitions, commercials, or invalid recordings. You can set the minimum duration of ' \
                   'a detected black interval above to adjust the sensitivity.</p>' \
                   '<p><b>WARNING:</b> this can take a long time to complete depending on the length and quality ' \
                   'of the source media.</p>'
            d = VCDoubleInputDialog(self, 'BLACKDETECT - Filter settings', 'Minimum duration for black scenes:',
                                    self.filter_settings.blackdetect.default_duration,
                                    self.filter_settings.blackdetect.min_duration, 999.9, 1, 0.1, desc, 'secs')
            d.buttons.accepted.connect(
                lambda: self.startFilters('detecting scenes (press ESC to cancel)',
                                          partial(self.videoService.blackdetect, d.value), d))
            d.setFixedSize(435, d.sizeHint().height())
            d.exec_()

    @pyqtSlot(str, partial, QDialog)
    def startFilters(self, progress_text: str, filter_func: partial, config_dialog: QDialog) -> None:
        config_dialog.close()
        self.parent.lock_gui(True)
        self.filterProgress(progress_text)
        filter_func()

    @pyqtSlot()
    def stopFilters(self) -> None:
        self.videoService.killFilterProc()
        self.parent.lock_gui(False)

    def filterProgress(self, msg: str) -> None:
        self.filterProgressBar = VCProgressDialog(self, modal=False)
        self.filterProgressBar.finished.connect(self.stopFilters)
        self.filterProgressBar.setText(msg)
        self.filterProgressBar.setMinimumWidth(600)
        self.filterProgressBar.show()

    @pyqtSlot()
    def addExternalClips(self) -> None:
        clips, _ = QFileDialog.getOpenFileNames(
            parent=self.parent,
            caption='Add media files',
            filter=self.mediaFilters(),
            initialFilter=self.mediaFilters(True),
            directory=(self.lastFolder if os.path.exists(self.lastFolder) else QDir.homePath()),
            options=self.getFileDialogOptions())
        if clips is not None and len(clips):
            self.lastFolder = QFileInfo(clips[0]).absolutePath()
            filesadded = False
            cliperrors = list()
            for file in clips:
                if len(self.clipTimes) > 0:
                    lastItem = self.clipTimes[len(self.clipTimes) - 1]
                    file4Test = lastItem[3] if len(lastItem[3]) else self.currentMedia
                    if self.videoService.testJoin(file4Test, file):
                        self.clipTimes.append([QTime(0, 0), self.videoService.duration(file),
                                               self.captureImage(file, QTime(0, 0, second=2), True), file])
                        filesadded = True
                    else:
                        cliperrors.append((file,
                                           (self.videoService.lastError if len(self.videoService.lastError) else '')))
                        self.videoService.lastError = ''
                else:
                    self.clipTimes.append([QTime(0, 0), self.videoService.duration(file),
                                           self.captureImage(file, QTime(0, 0, second=2), True), file])
                    filesadded = True
            if len(cliperrors):
                detailedmsg = '''<p>The file(s) listed were found to be incompatible for inclusion to the clip index as
                            they failed to join in simple tests used to ensure their compatibility. This is
                            commonly due to differences in frame size, audio/video formats (codecs), or both.</p>
                            <p>You can join these files as they currently are using traditional video editors like
                            OpenShot, Kdenlive, ShotCut, Final Cut Pro or Adobe Premiere. They can re-encode media
                            files with mixed properties so that they are then matching and able to be joined but
                            be aware that this can be a time consuming process and almost always results in
                            degraded video quality.</p>
                            <p>Re-encoding video is not going to ever be supported by VidCutter because those tools
                            are already available for you both free and commercially.</p>'''
                errordialog = ClipErrorsDialog(cliperrors, self)
                errordialog.setDetailedMessage(detailedmsg)
                errordialog.show()
            if filesadded:
                self.showText('media added to index')
                self.renderClipIndex()

    def hasExternals(self) -> bool:
        return True in [len(item[3]) > 0 for item in self.clipTimes]

    def clipStart(self) -> None:
        starttime = self.delta2QTime(self.seekSlider.value())
        self.clipTimes.append([starttime, '', self.captureImage(self.currentMedia, starttime), '', None])
        self.timeCounter.setMinimum(starttime.toString(self.timeformat))
        self.frameCounter.lockMinimum()
        self.toolbar_start.setDisabled(True)
        self.toolbar_end.setEnabled(True)
        self.clipindex_add.setDisabled(True)
        self.seekSlider.setRestrictValue(self.seekSlider.value(), True)
        self.blackdetectAction.setDisabled(True)
        self.inCut = True
        self.showText('clip started at {}'.format(starttime.toString(self.timeformat)))
        self.renderClipIndex()
        self.cliplist.scrollToBottom()

    def clipEnd(self) -> None:
        item = self.clipTimes[len(self.clipTimes) - 1]
        endtime = self.delta2QTime(self.seekSlider.value())
        if endtime.__lt__(item[0]):
            QMessageBox.critical(self.parent, 'Invalid END Time',
                                 'The clip end time must come AFTER it\'s start time. Please try again.')
            return
        item[1] = endtime
        self.toolbar_start.setEnabled(True)
        self.toolbar_end.setDisabled(True)
        self.clipindex_add.setEnabled(True)
        self.timeCounter.setMinimum()
        self.seekSlider.setRestrictValue(0, False)
        self.blackdetectAction.setEnabled(True)
        self.inCut = False
        self.showText('clip ends at {}'.format(endtime.toString(self.timeformat)))
        self.renderClipIndex()
        self.cliplist.scrollToBottom()

    @pyqtSlot()
    @pyqtSlot(bool)
    def setProjectDirty(self, dirty: bool=True) -> None:
        self.projectDirty = dirty

    # noinspection PyUnusedLocal,PyUnusedLocal,PyUnusedLocal
    @pyqtSlot(QModelIndex, int, int, QModelIndex, int)
    def syncClipList(self, parent: QModelIndex, start: int, end: int, destination: QModelIndex, row: int) -> None:
        index = row - 1 if start < row else row
        clip = self.clipTimes.pop(start)
        self.clipTimes.insert(index, clip)
        if not len(clip[3]):
            self.seekSlider.switchRegions(start, index)
        self.showText('clip order updated')
        self.renderClipIndex()

    def renderClipIndex(self) -> None:
        self.seekSlider.clearRegions()
        self.totalRuntime = 0
        externals = self.cliplist.renderClips(self.clipTimes)
        if len(self.clipTimes) and not self.inCut and externals != 1:
            self.toolbar_save.setEnabled(True)
            self.saveProjectAction.setEnabled(True)
        if self.inCut or len(self.clipTimes) == 0 or not isinstance(self.clipTimes[0][1], QTime):
            self.toolbar_save.setEnabled(False)
            self.saveProjectAction.setEnabled(False)
        self.setRunningTime(self.delta2QTime(self.totalRuntime).toString(self.runtimeformat))

    @staticmethod
    def delta2QTime(msecs: Union[float, int]) -> QTime:
        if isinstance(msecs, float):
            msecs = round(msecs * 1000)
        t = QTime(0, 0)
        return t.addMSecs(msecs)

    @staticmethod
    def qtime2delta(qtime: QTime) -> float:
        return timedelta(hours=qtime.hour(), minutes=qtime.minute(), seconds=qtime.second(),
                         milliseconds=qtime.msec()).total_seconds()

    @staticmethod
    def delta2String(td: timedelta) -> str:
        if td is None or td == timedelta.max:
            return ''
        else:
            return '%f' % (td.days * 86400 + td.seconds + td.microseconds / 1000000.)

    def captureImage(self, source: str, frametime: QTime, external: bool = False) -> QPixmap:
        return VideoService.captureFrame(self.settings, source, frametime.toString(self.timeformat), external=external)

    def saveMedia(self) -> None:
        clips = len(self.clipTimes)
        source_file, source_ext = os.path.splitext(self.currentMedia if self.currentMedia is not None
                                                   else self.clipTimes[0][3])
        suggestedFilename = '{0}_EDIT{1}'.format(source_file, source_ext)
        filefilter = 'Video files (*{0})'.format(source_ext)
        if clips > 0:
            self.finalFilename, _ = QFileDialog.getSaveFileName(
                parent=self.parent,
                caption='Save media file',
                directory=suggestedFilename,
                filter=filefilter,
                options=self.getFileDialogOptions())
            if self.finalFilename is None or not len(self.finalFilename.strip()):
                return
            file, ext = os.path.splitext(self.finalFilename)
            if len(ext) == 0 and len(source_ext):
                self.finalFilename += source_ext
            self.lastFolder = QFileInfo(self.finalFilename).absolutePath()
            self.toolbar_save.setDisabled(True)
            if not os.path.isdir(self.workFolder):
                os.mkdir(self.workFolder)
            if self.smartcut:
                self.seekSlider.showProgress(6 if clips > 1 else 5)
                self.parent.lock_gui(True)
                self.videoService.smartinit(clips)
                self.smartcutter(file, source_file, source_ext)
                return
            steps = 3 if clips > 1 else 2
            self.seekSlider.showProgress(steps)
            self.parent.lock_gui(True)
            filename, filelist = '', []
            for index, clip in enumerate(self.clipTimes):
                self.seekSlider.updateProgress(index)
                if len(clip[3]):
                    filelist.append(clip[3])
                else:
                    duration = self.delta2QTime(clip[0].msecsTo(clip[1])).toString(self.timeformat)
                    filename = '{0}_{1}{2}'.format(file, '{0:0>2}'.format(index), source_ext)
                    if not self.keepClips:
                        filename = os.path.join(self.workFolder, os.path.basename(filename))
                    filename = QDir.toNativeSeparators(filename)
                    filelist.append(filename)
                    if not self.videoService.cut(source='{0}{1}'.format(source_file, source_ext),
                                                 output=filename,
                                                 frametime=clip[0].toString(self.timeformat),
                                                 duration=duration,
                                                 allstreams=True):
                        self.completeOnError('<p>Failed to cut media file, assuming media is invalid or corrupt. '
                                             'Attempts are made to work around problematic media files, even '
                                             'when keyframes are incorrectly set or missing.</p><p>If you feel this '
                                             'is a bug in the software then please take the time to report it '
                                             'at our <a href="{}">GitHub Issues page</a> so that it can be fixed.</p>'
                                             .format(vidcutter.__bugreport__))
                        return
            self.joinMedia(filelist)

    def smartcutter(self, file: str, source_file: str, source_ext: str) -> None:
        self.smartcut_monitor = Munch(clips=[], results=[], externals=0)
        for index, clip in enumerate(self.clipTimes):
            if len(clip[3]):
                self.smartcut_monitor.clips.append(clip[3])
                self.smartcut_monitor.externals += 1
                if index == len(self.clipTimes):
                    self.smartmonitor()
            else:
                filename = '{0}_{1}{2}'.format(file, '{0:0>2}'.format(index), source_ext)
                if not self.keepClips:
                    filename = os.path.join(self.workFolder, os.path.basename(filename))
                filename = QDir.toNativeSeparators(filename)
                self.smartcut_monitor.clips.append(filename)
                self.videoService.smartcut(index=index,
                                           source='{0}{1}'.format(source_file, source_ext),
                                           output=filename,
                                           start=VideoCutter.qtime2delta(clip[0]),
                                           end=VideoCutter.qtime2delta(clip[1]),
                                           allstreams=True)

    @pyqtSlot(bool, str)
    def smartmonitor(self, success: bool = None, outputfile: str = None) -> None:
        if success is not None:
            if not success:
                self.logger.error('SmartCut failed for {}'.format(outputfile))
            self.smartcut_monitor.results.append(success)
        if len(self.smartcut_monitor.results) == len(self.smartcut_monitor.clips) - self.smartcut_monitor.externals:
            if False not in self.smartcut_monitor.results:
                self.joinMedia(self.smartcut_monitor.clips)

    def joinMedia(self, filelist: list) -> None:
        if len(filelist) > 1:
            self.seekSlider.updateProgress()
            rc = False
            chapters = None
            if self.createChapters:
                chapters = []
                [
                    chapters.append(clip[4] if clip[4] is not None else 'Chapter {}'.format(index + 1))
                    for index, clip in enumerate(self.clipTimes)
                ]
            if self.videoService.isMPEGcodec(filelist[0]):
                self.logger.info('source file is MPEG based so join via MPEG-TS')
                rc = self.videoService.mpegtsJoin(filelist, self.finalFilename, chapters)
            if not rc or QFile(self.finalFilename).size() < 1000:
                self.logger.info('MPEG-TS based join failed, will retry using standard concat')
                rc = self.videoService.join(filelist, self.finalFilename, True, chapters)
            if not rc or QFile(self.finalFilename).size() < 1000:
                self.logger.info('join resulted in 0 length file, trying again without all stream mapping')
                self.videoService.join(filelist, self.finalFilename, False, chapters)
            if not self.keepClips:
                for f in filelist:
                    clip = self.clipTimes[filelist.index(f)]
                    if not len(clip[3]) and os.path.isfile(f):
                        QFile.remove(f)
            self.complete(False)
        else:
            self.complete(True, filelist[-1])

    def complete(self, rename: bool=True, filename: str=None) -> None:
        if rename and filename is not None:
            # noinspection PyCallByClass
            QFile.remove(self.finalFilename)
            # noinspection PyCallByClass
            QFile.rename(filename, self.finalFilename)
        self.videoService.finalize(self.finalFilename)
        self.seekSlider.updateProgress()
        self.toolbar_save.setEnabled(True)
        self.parent.lock_gui(False)
        self.notify = JobCompleteNotification(
            self.finalFilename,
            self.sizeof_fmt(int(QFileInfo(self.finalFilename).size())),
            self.delta2QTime(self.totalRuntime).toString(self.runtimeformat),
            self.getAppIcon(encoded=True),
            self)
        self.notify.closed.connect(self.seekSlider.clearProgress)
        self.notify.show()
        if self.smartcut:
            QTimer.singleShot(1000, self.cleanup)
        self.setProjectDirty(False)

    @pyqtSlot(str)
    def completeOnError(self, errormsg: str) -> None:
        if self.smartcut:
            self.videoService.smartabort()
            QTimer.singleShot(1500, self.cleanup)
        self.parent.lock_gui(False)
        self.seekSlider.clearProgress()
        self.toolbar_save.setEnabled(True)
        self.parent.errorHandler(errormsg)

    def cleanup(self) -> None:
        if hasattr(self.videoService, 'smartcut_jobs'):
            delattr(self.videoService, 'smartcut_jobs')
        if hasattr(self, 'smartcut_monitor'):
            delattr(self, 'smartcut_monitor')
        self.videoService.smartcutError = False

    def saveSetting(self, setting: str, checked: bool) -> None:
        self.settings.setValue(setting, 'on' if checked else 'off')

    @pyqtSlot()
    def mediaInfo(self) -> None:
        if self.mediaAvailable:
            if self.videoService.backends.mediainfo is None:
                self.logger.error('mediainfo could not be found on the system')
                QMessageBox.critical(self.parent, 'Missing mediainfo utility',
                                     'The <b>mediainfo</b> command could not be found on your system which '
                                     'is required for this feature to work.<br/><br/>Linux users can simply '
                                     'install the <b>mediainfo</b> package using the package manager you use to '
                                     'install software (e.g. apt, pacman, dnf, zypper, etc.)')
                return
            mediainfo = MediaInfo(media=self.currentMedia, parent=self)
            mediainfo.show()

    @pyqtSlot()
    def selectStreams(self) -> None:
        if self.mediaAvailable and self.videoService.streams:
            if self.hasExternals():
                nostreamstext = '''
                    <style>
                        h2 {{
                            color: {0};
                            font-family: "Futura LT", sans-serif;
                            font-weight: normal;
                        }}
                    </style>
                    <table border="0" cellpadding="6" cellspacing="0" width="350">
                        <tr>
                            <td><h2>Cannot configure stream selection</h2></td>
                        </tr>
                        <tr>
                            <td>
                                Stream selection cannot be configured when external media files
                                are added to your clip index. Remove all external files from your
                                clip index and try again.
                            </td>
                        </tr>
                    </table>'''.format('#C681D5' if self.theme == 'dark' else '#642C68')
                nostreams = QMessageBox(QMessageBox.Critical,
                                        'Stream selection is unavailable',
                                        nostreamstext,
                                        parent=self.parent)
                nostreams.setStandardButtons(QMessageBox.Ok)
                nostreams.exec_()
                return
            streamSelector = StreamSelector(self.videoService, self)
            streamSelector.show()

    def saveWarning(self) -> tuple:
        if self.mediaAvailable and self.projectDirty and not self.projectSaved:
            savewarn = VCMessageBox('Warning', 'Unsaved changes found in project',
                                    'Would you like to save your project?', parent=self)
            savebutton = savewarn.addButton('Save project', QMessageBox.YesRole)
            savewarn.addButton('Do not save', QMessageBox.NoRole)
            cancelbutton = savewarn.addButton('Cancel', QMessageBox.RejectRole)
            savewarn.exec_()
            res = savewarn.clickedButton()
            if res == savebutton:
                return True, self.saveProject
            elif res == cancelbutton:
                return True, None
        return False, None

    @pyqtSlot()
    def showKeyRef(self) -> None:
        msgtext = '<img src=":/images/{}/shortcuts.png" />'.format(self.theme)
        msgbox = QMessageBox(QMessageBox.NoIcon, 'Keyboard shortcuts', msgtext, QMessageBox.Ok, self,
                             Qt.Window | Qt.Dialog | Qt.WindowCloseButtonHint)
        msgbox.setObjectName('shortcuts')
        msgbox.setContentsMargins(10, 10, 10, 10)
        msgbox.setMinimumWidth(400 if self.parent.scale == 'LOW' else 600)
        msgbox.exec_()

    @pyqtSlot()
    def aboutApp(self) -> None:
        about = About(self.videoService, self.mpvWidget, self)
        about.exec_()

    @staticmethod
    def getAppIcon(encoded: bool=False):
        icon = QIcon.fromTheme(qApp.applicationName().lower(), QIcon(':/images/vidcutter-small.png'))
        if not encoded:
            return icon
        iconimg = icon.pixmap(82, 82).toImage()
        data = QByteArray()
        buffer = QBuffer(data)
        buffer.open(QBuffer.WriteOnly)
        iconimg.save(buffer, 'PNG')
        base64enc = str(data.toBase64().data(), 'latin1')
        icon = 'data:vidcutter.png;base64,{}'.format(base64enc)
        return icon

    @staticmethod
    def sizeof_fmt(num: float, suffix: chr='B') -> str:
        for unit in ['', 'K', 'M', 'G', 'T', 'P', 'E', 'Z']:
            if abs(num) < 1024.0:
                return "%3.1f %s%s" % (num, unit, suffix)
            num /= 1024.0
        return "%.1f %s%s" % (num, 'Y', suffix)

    @pyqtSlot()
    def viewChangelog(self) -> None:
        changelog = Changelog(self)
        changelog.exec_()

    @staticmethod
    @pyqtSlot()
    def viewLogs() -> None:
        QDesktopServices.openUrl(QUrl.fromLocalFile(logging.getLoggerClass().root.handlers[0].baseFilename))

    @pyqtSlot()
    def toggleFullscreen(self) -> None:
        if self.mediaAvailable:
            pause = self.mpvWidget.property('pause')
            mute = self.mpvWidget.property('mute')
            vol = self.mpvWidget.property('volume')
            pos = self.seekSlider.value() / 1000
            if self.mpvWidget.originalParent is not None:
                self.mpvWidget.shutdown()
                sip.delete(self.mpvWidget)
                del self.mpvWidget
                self.mpvWidget = self.getMPV(parent=self, file=self.currentMedia, start=pos, pause=pause, mute=mute,
                                             volume=vol)
                self.videoplayerLayout.insertWidget(0, self.mpvWidget)
                self.mpvWidget.originalParent = None
                self.parent.show()
            elif self.mpvWidget.parentWidget() != 0:
                self.parent.hide()
                self.mpvWidget.shutdown()
                self.videoplayerLayout.removeWidget(self.mpvWidget)
                sip.delete(self.mpvWidget)
                del self.mpvWidget
                self.mpvWidget = self.getMPV(file=self.currentMedia, start=pos, pause=pause, mute=mute, volume=vol)
                self.mpvWidget.originalParent = self
                self.mpvWidget.setGeometry(qApp.desktop().screenGeometry(self))
                self.mpvWidget.showFullScreen()

    def toggleOSD(self, checked: bool) -> None:
        self.showText('on-screen display {}'.format('enabled' if checked else 'disabled'), override=True)
        self.saveSetting('enableOSD', checked)

    @property
    def _osdfont(self) -> str:
        fontdb = QFontDatabase()
        return 'DejaVu Sans' if 'DejaVu Sans' in fontdb.families(QFontDatabase.Latin) else 'Noto Sans'

    def doPass(self) -> None:
        pass

    def keyPressEvent(self, event: QKeyEvent) -> None:
        if event.key() in {Qt.Key_Q, Qt.Key_W} and event.modifiers() == Qt.ControlModifier:
            self.parent.close()
            return

        if self.mediaAvailable:

            if event.key() == Qt.Key_Space:
                self.playMedia()
                return

            if event.key() == Qt.Key_Escape and self.isFullScreen():
                self.toggleFullscreen()
                return

            if event.key() == Qt.Key_F:
                self.toggleFullscreen()
                return

            if event.key() == Qt.Key_Home:
                self.setPosition(self.seekSlider.minimum())
                return

            if event.key() == Qt.Key_End:
                self.setPosition(self.seekSlider.maximum())
                return

            if event.key() == Qt.Key_Left:
                self.mpvWidget.frameBackStep()
                self.setPlayButton(False)
                return

            if event.key() == Qt.Key_Down:
                if qApp.queryKeyboardModifiers() == Qt.ShiftModifier:
                    self.mpvWidget.seek(-self.level2Seek, 'relative+exact')
                else:
                    self.mpvWidget.seek(-self.level1Seek, 'relative+exact')
                return

            if event.key() == Qt.Key_Right:
                self.mpvWidget.frameStep()
                self.setPlayButton(False)
                return

            if event.key() == Qt.Key_Up:
                if qApp.queryKeyboardModifiers() == Qt.ShiftModifier:
                    self.mpvWidget.seek(self.level2Seek, 'relative+exact')
                else:
                    self.mpvWidget.seek(self.level1Seek, 'relative+exact')
                return

            if event.key() in {Qt.Key_Return, Qt.Key_Enter} and \
                    (not self.timeCounter.hasFocus() and not self.frameCounter.hasFocus()):
                if self.toolbar_start.isEnabled():
                    self.clipStart()
                elif self.toolbar_end.isEnabled():
                    self.clipEnd()
                return

        super(VideoCutter, self).keyPressEvent(event)

    def showEvent(self, event: QShowEvent) -> None:
        if hasattr(self, 'filterProgressBar') and self.filterProgressBar.isVisible():
            self.filterProgressBar.update()
        super(VideoCutter, self).showEvent(event)



# -*- coding: utf-8 -*-
"""
ISO639-2 standart: the module only contains the dictionary ISO639_2
which maps a language code in three letters (eg. "fre") to a language
name in english (eg. "French").
"""

# ISO-639, the list comes from:
# http://www.loc.gov/standards/iso639-2/php/English_list.php
_ISO639 = (
    (u"Abkhazian", "abk", "ab"),
    (u"Achinese", "ace", None),
    (u"Acoli", "ach", None),
    (u"Adangme", "ada", None),
    (u"Adygei", "ady", None),
    (u"Adyghe", "ady", None),
    (u"Afar", "aar", "aa"),
    (u"Afrihili", "afh", None),
    (u"Afrikaans", "afr", "af"),
    (u"Afro-Asiatic (Other)", "afa", None),
    (u"Ainu", "ain", None),
    (u"Akan", "aka", "ak"),
    (u"Akkadian", "akk", None),
    (u"Albanian", "alb/sqi", "sq"),
    (u"Alemani", "gsw", None),
    (u"Aleut", "ale", None),
    (u"Algonquian languages", "alg", None),
    (u"Altaic (Other)", "tut", None),
    (u"Amharic", "amh", "am"),
    (u"Angika", "anp", None),
    (u"Apache languages", "apa", None),
    (u"Arabic", "ara", "ar"),
    (u"Aragonese", "arg", "an"),
    (u"Aramaic", "arc", None),
    (u"Arapaho", "arp", None),
    (u"Araucanian", "arn", None),
    (u"Arawak", "arw", None),
    (u"Armenian", "arm/hye", "hy"),
    (u"Aromanian", "rup", None),
    (u"Artificial (Other)", "art", None),
    (u"Arumanian", "rup", None),
    (u"Assamese", "asm", "as"),
    (u"Asturian", "ast", None),
    (u"Athapascan languages", "ath", None),
    (u"Australian languages", "aus", None),
    (u"Austronesian (Other)", "map", None),
    (u"Avaric", "ava", "av"),
    (u"Avestan", "ave", "ae"),
    (u"Awadhi", "awa", None),
    (u"Aymara", "aym", "ay"),
    (u"Azerbaijani", "aze", "az"),
    (u"Bable", "ast", None),
    (u"Balinese", "ban", None),
    (u"Baltic (Other)", "bat", None),
    (u"Baluchi", "bal", None),
    (u"Bambara", "bam", "bm"),
    (u"Bamileke languages", "bai", None),
    (u"Banda", "bad", None),
    (u"Bantu (Other)", "bnt", None),
    (u"Basa", "bas", None),
    (u"Bashkir", "bak", "ba"),
    (u"Basque", "baq/eus", "eu"),
    (u"Batak (Indonesia)", "btk", None),
    (u"Beja", "bej", None),
    (u"Belarusian", "bel", "be"),
    (u"Bemba", "bem", None),
    (u"Bengali", "ben", "bn"),
    (u"Berber (Other)", "ber", None),
    (u"Bhojpuri", "bho", None),
    (u"Bihari", "bih", "bh"),
    (u"Bikol", "bik", None),
    (u"Bilin", "byn", None),
    (u"Bini", "bin", None),
    (u"Bislama", "bis", "bi"),
    (u"Blin", "byn", None),
    (u"Bokmål, Norwegian", "nob", "nb"),
    (u"Bosnian", "bos", "bs"),
    (u"Braj", "bra", None),
    (u"Breton", "bre", "br"),
    (u"Buginese", "bug", None),
    (u"Bulgarian", "bul", "bg"),
    (u"Buriat", "bua", None),
    (u"Burmese", "bur/mya", "my"),
    (u"Caddo", "cad", None),
    (u"Carib", "car", None),
    (u"Castilian", "spa", "es"),
    (u"Catalan", "cat", "ca"),
    (u"Caucasian (Other)", "cau", None),
    (u"Cebuano", "ceb", None),
    (u"Celtic (Other)", "cel", None),
    (u"Central American Indian (Other)", "cai", None),
    (u"Chagatai", "chg", None),
    (u"Chamic languages", "cmc", None),
    (u"Chamorro", "cha", "ch"),
    (u"Chechen", "che", "ce"),
    (u"Cherokee", "chr", None),
    (u"Chewa", "nya", "ny"),
    (u"Cheyenne", "chy", None),
    (u"Chibcha", "chb", None),
    (u"Chichewa", "nya", "ny"),
    (u"Chinese", "chi/zho", "zh"),
    (u"Chinook jargon", "chn", None),
    (u"Chipewyan", "chp", None),
    (u"Choctaw", "cho", None),
    (u"Chuang", "zha", "za"),
    (u"Church Slavic", "chu", "cu"),
    (u"Church Slavonic", "chu", "cu"),
    (u"Chuukese", "chk", None),
    (u"Chuvash", "chv", "cv"),
    (u"Classical Nepal Bhasa", "nwc", None),
    (u"Classical Newari", "nwc", None),
    (u"Coptic", "cop", None),
    (u"Cornish", "cor", "kw"),
    (u"Corsican", "cos", "co"),
    (u"Cree", "cre", "cr"),
    (u"Creek", "mus", None),
    (u"Creoles and pidgins (Other)", "crp", None),
    (u"Creoles and pidgins, English based (Other)", "cpe", None),
    (u"Creoles and pidgins, French-based (Other)", "cpf", None),
    (u"Creoles and pidgins, Portuguese-based (Other)", "cpp", None),
    (u"Crimean Tatar", "crh", None),
    (u"Crimean Turkish", "crh", None),
    (u"Croatian", "scr/hrv", "hr"),
    (u"Cushitic (Other)", "cus", None),
    (u"Czech", "cze/ces", "cs"),
    (u"Dakota", "dak", None),
    (u"Danish", "dan", "da"),
    (u"Dargwa", "dar", None),
    (u"Dayak", "day", None),
    (u"Delaware", "del", None),
    (u"Dhivehi", "div", "dv"),
    (u"Dimili", "zza", None),
    (u"Dimli", "zza", None),
    (u"Dinka", "din", None),
    (u"Divehi", "div", "dv"),
    (u"Dogri", "doi", None),
    (u"Dogrib", "dgr", None),
    (u"Dravidian (Other)", "dra", None),
    (u"Duala", "dua", None),
    (u"Dutch", "dut", "nl"),
    (u"Dutch", "nld", "nl"),
    (u"Dyula", "dyu", None),
    (u"Dzongkha", "dzo", "dz"),
    (u"Eastern Frisian", "frs", None),
    (u"Efik", "efi", None),
    (u"Egyptian (Ancient)", "egy", None),
    (u"Ekajuk", "eka", None),
    (u"Elamite", "elx", None),
    (u"English", "eng", "en"),
    (u"English, Middle (1100-1500)", "enm", None),
    (u"English, Old (ca.450-1100)", "ang", None),
    (u"Erzya", "myv", None),
    (u"Esperanto", "epo", "eo"),
    (u"Estonian", "est", "et"),
    (u"Ewe", "ewe", "ee"),
    (u"Ewondo", "ewo", None),
    (u"Fang", "fan", None),
    (u"Fanti", "fat", None),
    (u"Faroese", "fao", "fo"),
    (u"Fijian", "fij", "fj"),
    (u"Filipino", "fil", None),
    (u"Finnish", "fin", "fi"),
    (u"Finno-Ugrian (Other)", "fiu", None),
    (u"Fon", "fon", None),
    (u"French", "fre/fra", "fr"),
    (u"French, Middle (ca.1400-1600)", "frm", None),
    (u"French, Old (842-ca.1400)", "fro", None),
    (u"Friulian", "fur", None),
    (u"Fulah", "ful", "ff"),
    (u"Ga", "gaa", None),
    (u"Gaelic", "gla", "gd"),
    (u"Galician", "glg", "gl"),
    (u"Ganda", "lug", "lg"),
    (u"Gayo", "gay", None),
    (u"Gbaya", "gba", None),
    (u"Geez", "gez", None),
    (u"Georgian", "geo/kat", "ka"),
    (u"German", "ger/deu", "de"),
    (u"German, Low", "nds", None),
    (u"German, Middle High (ca.1050-1500)", "gmh", None),
    (u"German, Old High (ca.750-1050)", "goh", None),
    (u"Germanic (Other)", "gem", None),
    (u"Gikuyu", "kik", "ki"),
    (u"Gilbertese", "gil", None),
    (u"Gondi", "gon", None),
    (u"Gorontalo", "gor", None),
    (u"Gothic", "got", None),
    (u"Grebo", "grb", None),
    (u"Greek, Ancient (to 1453)", "grc", None),
    (u"Greek, Modern (1453-)", "gre/ell", "el"),
    (u"Greenlandic", "kal", "kl"),
    (u"Guarani", "grn", "gn"),
    (u"Gujarati", "guj", "gu"),
    (u"Gwich´in", "gwi", None),
    (u"Haida", "hai", None),
    (u"Haitian", "hat", "ht"),
    (u"Haitian Creole", "hat", "ht"),
    (u"Hausa", "hau", "ha"),
    (u"Hawaiian", "haw", None),
    (u"Hebrew", "heb", "he"),
    (u"Herero", "her", "hz"),
    (u"Hiligaynon", "hil", None),
    (u"Himachali", "him", None),
    (u"Hindi", "hin", "hi"),
    (u"Hiri Motu", "hmo", "ho"),
    (u"Hittite", "hit", None),
    (u"Hmong", "hmn", None),
    (u"Hungarian", "hun", "hu"),
    (u"Hupa", "hup", None),
    (u"Iban", "iba", None),
    (u"Icelandic", "ice/isl", "is"),
    (u"Ido", "ido", "io"),
    (u"Igbo", "ibo", "ig"),
    (u"Ijo", "ijo", None),
    (u"Iloko", "ilo", None),
    (u"Inari Sami", "smn", None),
    (u"Indic (Other)", "inc", None),
    (u"Indo-European (Other)", "ine", None),
    (u"Indonesian", "ind", "id"),
    (u"Ingush", "inh", None),
    (u"Interlingua", "ina", "ia"),
    (u"Interlingue", "ile", "ie"),
    (u"Inuktitut", "iku", "iu"),
    (u"Inupiaq", "ipk", "ik"),
    (u"Iranian (Other)", "ira", None),
    (u"Irish", "gle", "ga"),
    (u"Irish, Middle (900-1200)", "mga", None),
    (u"Irish, Old (to 900)", "sga", None),
    (u"Iroquoian languages", "iro", None),
    (u"Italian", "ita", "it"),
    (u"Japanese", "jpn", "ja"),
    (u"Javanese", "jav", "jv"),
    (u"Judeo-Arabic", "jrb", None),
    (u"Judeo-Persian", "jpr", None),
    (u"Kabardian", "kbd", None),
    (u"Kabyle", "kab", None),
    (u"Kachin", "kac", None),
    (u"Kalaallisut", "kal", "kl"),
    (u"Kalmyk", "xal", None),
    (u"Kamba", "kam", None),
    (u"Kannada", "kan", "kn"),
    (u"Kanuri", "kau", "kr"),
    (u"Kara-Kalpak", "kaa", None),
    (u"Karachay-Balkar", "krc", None),
    (u"Karelian", "krl", None),
    (u"Karen", "kar", None),
    (u"Kashmiri", "kas", "ks"),
    (u"Kashubian", "csb", None),
    (u"Kawi", "kaw", None),
    (u"Kazakh", "kaz", "kk"),
    (u"Khasi", "kha", None),
    (u"Khmer", "khm", "km"),
    (u"Khoisan (Other)", "khi", None),
    (u"Khotanese", "kho", None),
    (u"Kikuyu", "kik", "ki"),
    (u"Kimbundu", "kmb", None),
    (u"Kinyarwanda", "kin", "rw"),
    (u"Kirdki", "zza", None),
    (u"Kirghiz", "kir", "ky"),
    (u"Kirmanjki", "zza", None),
    (u"Klingon", "tlh", None),
    (u"Komi", "kom", "kv"),
    (u"Kongo", "kon", "kg"),
    (u"Konkani", "kok", None),
    (u"Korean", "kor", "ko"),
    (u"Kosraean", "kos", None),
    (u"Kpelle", "kpe", None),
    (u"Kru", "kro", None),
    (u"Kuanyama", "kua", "kj"),
    (u"Kumyk", "kum", None),
    (u"Kurdish", "kur", "ku"),
    (u"Kurukh", "kru", None),
    (u"Kutenai", "kut", None),
    (u"Kwanyama", "kua", "kj"),
    (u"Ladino", "lad", None),
    (u"Lahnda", "lah", None),
    (u"Lamba", "lam", None),
    (u"Lao", "lao", "lo"),
    (u"Latin", "lat", "la"),
    (u"Latvian", "lav", "lv"),
    (u"Letzeburgesch", "ltz", "lb"),
    (u"Lezghian", "lez", None),
    (u"Limburgan", "lim", "li"),
    (u"Limburger", "lim", "li"),
    (u"Limburgish", "lim", "li"),
    (u"Lingala", "lin", "ln"),
    (u"Lithuanian", "lit", "lt"),
    (u"Lojban", "jbo", None),
    (u"Low German", "nds", None),
    (u"Low Saxon", "nds", None),
    (u"Lower Sorbian", "dsb", None),
    (u"Lozi", "loz", None),
    (u"Luba-Katanga", "lub", "lu"),
    (u"Luba-Lulua", "lua", None),
    (u"Luiseno", "lui", None),
    (u"Lule Sami", "smj", None),
    (u"Lunda", "lun", None),
    (u"Luo (Kenya and Tanzania)", "luo", None),
    (u"Lushai", "lus", None),
    (u"Luxembourgish", "ltz", "lb"),
    (u"Macedo-Romanian", "rup", None),
    (u"Macedonian", "mac/mkd", "mk"),
    (u"Madurese", "mad", None),
    (u"Magahi", "mag", None),
    (u"Maithili", "mai", None),
    (u"Makasar", "mak", None),
    (u"Malagasy", "mlg", "mg"),
    (u"Malay", "may/msa", "ms"),
    (u"Malayalam", "mal", "ml"),
    (u"Maldivian", "div", "dv"),
    (u"Maltese", "mlt", "mt"),
    (u"Manchu", "mnc", None),
    (u"Mandar", "mdr", None),
    (u"Mandingo", "man", None),
    (u"Manipuri", "mni", None),
    (u"Manobo languages", "mno", None),
    (u"Manx", "glv", "gv"),
    (u"Maori", "mao/mri", "mi"),
    (u"Marathi", "mar", "mr"),
    (u"Mari", "chm", None),
    (u"Marshallese", "mah", "mh"),
    (u"Marwari", "mwr", None),
    (u"Masai", "mas", None),
    (u"Mayan languages", "myn", None),
    (u"Mende", "men", None),
    (u"Mi'kmaq", "mic", None),
    (u"Micmac", "mic", None),
    (u"Minangkabau", "min", None),
    (u"Mirandese", "mwl", None),
    (u"Miscellaneous languages", "mis", None),
    (u"Mohawk", "moh", None),
    (u"Moksha", "mdf", None),
    (u"Moldavian", "mol", "mo"),
    (u"Mon-Khmer (Other)", "mkh", None),
    (u"Mongo", "lol", None),
    (u"Mongolian", "mon", "mn"),
    (u"Mossi", "mos", None),
    (u"Multiple languages", "mul", None),
    (u"Munda languages", "mun", None),
    (u"N'Ko", "nqo", None),
    (u"Nahuatl", "nah", None),
    (u"Nauru", "nau", "na"),
    (u"Navaho", "nav", "nv"),
    (u"Navajo", "nav", "nv"),
    (u"Ndebele, North", "nde", "nd"),
    (u"Ndebele, South", "nbl", "nr"),
    (u"Ndonga", "ndo", "ng"),
    (u"Neapolitan", "nap", None),
    (u"Nepal Bhasa", "new", None),
    (u"Nepali", "nep", "ne"),
    (u"Newari", "new", None),
    (u"Nias", "nia", None),
    (u"Niger-Kordofanian (Other)", "nic", None),
    (u"Nilo-Saharan (Other)", "ssa", None),
    (u"Niuean", "niu", None),
    (u"No linguistic content", "zxx", None),
    (u"Nogai", "nog", None),
    (u"Norse, Old", "non", None),
    (u"North American Indian", "nai", None),
    (u"North Ndebele", "nde", "nd"),
    (u"Northern Frisian", "frr", None),
    (u"Northern Sami", "sme", "se"),
    (u"Northern Sotho", "nso", None),
    (u"Norwegian", "nor", "no"),
    (u"Norwegian Bokmål", "nob", "nb"),
    (u"Norwegian Nynorsk", "nno", "nn"),
    (u"Nubian languages", "nub", None),
    (u"Nyamwezi", "nym", None),
    (u"Nyanja", "nya", "ny"),
    (u"Nyankole", "nyn", None),
    (u"Nynorsk, Norwegian", "nno", "nn"),
    (u"Nyoro", "nyo", None),
    (u"Nzima", "nzi", None),
    (u"Occitan (post 1500)", "oci", "oc"),
    (u"Oirat", "xal", None),
    (u"Ojibwa", "oji", "oj"),
    (u"Old Bulgarian", "chu", "cu"),
    (u"Old Church Slavonic", "chu", "cu"),
    (u"Old Newari", "nwc", None),
    (u"Old Slavonic", "chu", "cu"),
    (u"Oriya", "ori", "or"),
    (u"Oromo", "orm", "om"),
    (u"Osage", "osa", None),
    (u"Ossetian", "oss", "os"),
    (u"Ossetic", "oss", "os"),
    (u"Otomian languages", "oto", None),
    (u"Pahlavi", "pal", None),
    (u"Palauan", "pau", None),
    (u"Pali", "pli", "pi"),
    (u"Pampanga", "pam", None),
    (u"Pangasinan", "pag", None),
    (u"Panjabi", "pan", "pa"),
    (u"Papiamento", "pap", None),
    (u"Papuan (Other)", "paa", None),
    (u"Pedi", "nso", None),
    (u"Persian", "per/fas", "fa"),
    (u"Persian, Old (ca.600-400 B.C.)", "peo", None),
    (u"Philippine (Other)", "phi", None),
    (u"Phoenician", "phn", None),
    (u"Pilipino", "fil", None),
    (u"Pohnpeian", "pon", None),
    (u"Polish", "pol", "pl"),
    (u"Portuguese", "por", "pt"),
    (u"Prakrit languages", "pra", None),
    (u"Provençal", "oci", "oc"),
    (u"Provençal, Old (to 1500)", "pro", None),
    (u"Punjabi", "pan", "pa"),
    (u"Pushto", "pus", "ps"),
    (u"Quechua", "que", "qu"),
    (u"Raeto-Romance", "roh", "rm"),
    (u"Rajasthani", "raj", None),
    (u"Rapanui", "rap", None),
    (u"Rarotongan", "rar", None),
    (u"Reserved for local use", "qaa/qtz", None),
    (u"Romance (Other)", "roa", None),
    (u"Romanian", "rum/ron", "ro"),
    (u"Romany", "rom", None),
    (u"Rundi", "run", "rn"),
    (u"Russian", "rus", "ru"),
    (u"Salishan languages", "sal", None),
    (u"Samaritan Aramaic", "sam", None),
    (u"Sami languages (Other)", "smi", None),
    (u"Samoan", "smo", "sm"),
    (u"Sandawe", "sad", None),
    (u"Sango", "sag", "sg"),
    (u"Sanskrit", "san", "sa"),
    (u"Santali", "sat", None),
    (u"Sardinian", "srd", "sc"),
    (u"Sasak", "sas", None),
    (u"Saxon, Low", "nds", None),
    (u"Scots", "sco", None),
    (u"Scottish Gaelic", "gla", "gd"),
    (u"Selkup", "sel", None),
    (u"Semitic (Other)", "sem", None),
    (u"Sepedi", "nso", None),
    (u"Serbian", "scc/srp", "sr"),
    (u"Serer", "srr", None),
    (u"Shan", "shn", None),
    (u"Shona", "sna", "sn"),
    (u"Sichuan Yi", "iii", "ii"),
    (u"Sicilian", "scn", None),
    (u"Sidamo", "sid", None),
    (u"Sign Languages", "sgn", None),
    (u"Siksika", "bla", None),
    (u"Sindhi", "snd", "sd"),
    (u"Sinhala", "sin", "si"),
    (u"Sinhalese", "sin", "si"),
    (u"Sino-Tibetan (Other)", "sit", None),
    (u"Siouan languages", "sio", None),
    (u"Skolt Sami", "sms", None),
    (u"Slave (Athapascan)", "den", None),
    (u"Slavic (Other)", "sla", None),
    (u"Slovak", "slo/slk", "sk"),
    (u"Slovenian", "slv", "sl"),
    (u"Sogdian", "sog", None),
    (u"Somali", "som", "so"),
    (u"Songhai", "son", None),
    (u"Soninke", "snk", None),
    (u"Sorbian languages", "wen", None),
    (u"Sotho, Northern", "nso", None),
    (u"Sotho, Southern", "sot", "st"),
    (u"South American Indian (Other)", "sai", None),
    (u"South Ndebele", "nbl", "nr"),
    (u"Southern Altai", "alt", None),
    (u"Southern Sami", "sma", None),
    (u"Spanish", "spa", "es"),
    (u"Sranan Togo", "srn", None),
    (u"Sukuma", "suk", None),
    (u"Sumerian", "sux", None),
    (u"Sundanese", "sun", "su"),
    (u"Susu", "sus", None),
    (u"Swahili", "swa", "sw"),
    (u"Swati", "ssw", "ss"),
    (u"Swedish", "swe", "sv"),
    (u"Swiss German", "gsw", None),
    (u"Syriac", "syr", None),
    (u"Tagalog", "tgl", "tl"),
    (u"Tahitian", "tah", "ty"),
    (u"Tai (Other)", "tai", None),
    (u"Tajik", "tgk", "tg"),
    (u"Tamashek", "tmh", None),
    (u"Tamil", "tam", "ta"),
    (u"Tatar", "tat", "tt"),
    (u"Telugu", "tel", "te"),
    (u"Tereno", "ter", None),
    (u"Tetum", "tet", None),
    (u"Thai", "tha", "th"),
    (u"Tibetan", "tib/bod", "bo"),
    (u"Tigre", "tig", None),
    (u"Tigrinya", "tir", "ti"),
    (u"Timne", "tem", None),
    (u"Tiv", "tiv", None),
    (u"tlhIngan-Hol", "tlh", None),
    (u"Tlingit", "tli", None),
    (u"Tok Pisin", "tpi", None),
    (u"Tokelau", "tkl", None),
    (u"Tonga (Nyasa)", "tog", None),
    (u"Tonga (Tonga Islands)", "ton", "to"),
    (u"Tsimshian", "tsi", None),
    (u"Tsonga", "tso", "ts"),
    (u"Tswana", "tsn", "tn"),
    (u"Tumbuka", "tum", None),
    (u"Tupi languages", "tup", None),
    (u"Turkish", "tur", "tr"),
    (u"Turkish, Ottoman (1500-1928)", "ota", None),
    (u"Turkmen", "tuk", "tk"),
    (u"Tuvalu", "tvl", None),
    (u"Tuvinian", "tyv", None),
    (u"Twi", "twi", "tw"),
    (u"Udmurt", "udm", None),
    (u"Ugaritic", "uga", None),
    (u"Uighur", "uig", "ug"),
    (u"Ukrainian", "ukr", "uk"),
    (u"Umbundu", "umb", None),
    (u"Undetermined", "und", None),
    (u"Upper Sorbian", "hsb", None),
    (u"Urdu", "urd", "ur"),
    (u"Uyghur", "uig", "ug"),
    (u"Uzbek", "uzb", "uz"),
    (u"Vai", "vai", None),
    (u"Valencian", "cat", "ca"),
    (u"Venda", "ven", "ve"),
    (u"Vietnamese", "vie", "vi"),
    (u"Volapük", "vol", "vo"),
    (u"Votic", "vot", None),
    (u"Wakashan languages", "wak", None),
    (u"Walamo", "wal", None),
    (u"Walloon", "wln", "wa"),
    (u"Waray", "war", None),
    (u"Washo", "was", None),
    (u"Welsh", "wel/cym", "cy"),
    (u"Western Frisian", "fry", "fy"),
    (u"Wolof", "wol", "wo"),
    (u"Xhosa", "xho", "xh"),
    (u"Yakut", "sah", None),
    (u"Yao", "yao", None),
    (u"Yapese", "yap", None),
    (u"Yiddish", "yid", "yi"),
    (u"Yoruba", "yor", "yo"),
    (u"Yupik languages", "ypk", None),
    (u"Zande", "znd", None),
    (u"Zapotec", "zap", None),
    (u"Zaza", "zza", None),
    (u"Zazaki", "zza", None),
    (u"Zenaga", "zen", None),
    (u"Zhuang", "zha", "za"),
    (u"Zulu", "zul", "zu"),
    (u"Zuni", "zun", None),
)

# Bibliographic ISO-639-2 form (eg. "fre" => "French")
ISO639_2 = {}
for line in _ISO639:
    for key in line[1].split("/"):
        ISO639_2[key] = line[0]
del _ISO639




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import logging
import math
import os
import sys

from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QCloseEvent, QShowEvent
from PyQt5.QtWidgets import (QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QPushButton, QSizePolicy, QStyleFactory,
                             QTextBrowser, QVBoxLayout, qApp)


class MediaInfo(QDialog):
    modes = {
        'LOW': QSize(450, 300),
        'NORMAL': QSize(600, 450),
        'HIGH': QSize(1080, 700)
    }

    def __init__(self, media, parent=None, flags=Qt.Dialog | Qt.WindowCloseButtonHint):
        super(MediaInfo, self).__init__(parent, flags)
        self.logger = logging.getLogger(__name__)
        self.media = media
        self.parent = parent
        self.setObjectName('mediainfo')
        self.setContentsMargins(0, 0, 0, 0)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle('Media information - {}'.format(os.path.basename(self.media)))
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.setMinimumSize(self.modes.get(self.parent.parent.scale))
        if sys.platform in {'win32', 'darwin'}:
            self.setStyle(QStyleFactory.create('Fusion'))
        metadata = '''<style>
    table {{
        font-family: "Noto Sans", sans-serif;
        font-size: 13px;
        border: 1px solid #999;
    }}
    td i {{
        font-family: "Noto Sans", sans-serif;
        font-weight: normal;
        font-style: normal;
        text-align: right;
        color: {pencolor};
        white-space: nowrap;
    }}
    td {{
        font-weight: normal;
        text-align: right;
    }}
    td + td {{ text-align: left; }}
    h1, h2, h3 {{ color: {pencolor}; }}
</style>
<div align="center">{info}</div>'''.format(pencolor='#C681D5' if self.parent.theme == 'dark' else '#642C68',
                                           info=self.parent.videoService.mediainfo(self.media))
        content = QTextBrowser(self.parent)
        if sys.platform in {'win32', 'darwin'}:
            content.setStyle(QStyleFactory.create('Fusion'))
        content.setStyleSheet('''
            QTextBrowser {{
                background-color: {bgcolor};
                color: {pencolor};
                border: 1px solid #999;
            }}'''.format(bgcolor='rgba(12, 15, 16, 210)' if self.parent.theme == 'dark' else 'rgba(255, 255, 255, 150)',
                         pencolor='#FFF' if self.parent.theme == 'dark' else '#000'))
        content.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        content.setHtml(metadata)
        keyframesButton = QPushButton('View keyframes', self)
        keyframesButton.clicked.connect(self.showKeyframes)
        okButton = QDialogButtonBox(QDialogButtonBox.Ok)
        okButton.accepted.connect(self.close)
        button_layout = QHBoxLayout()
        mediainfo_version = self.parent.videoService.cmdExec(self.parent.videoService.backends.mediainfo,
                                                             '--version', True)
        if len(mediainfo_version) >= 2:
            mediainfo_version = mediainfo_version.split('\n')[1]
            mediainfo_label = QLabel('<div style="font-size:11px;"><b>Media information by:</b><br/>%s @ '
                                     % mediainfo_version + '<a href="https://mediaarea.net" target="_blank">' +
                                     'mediaarea.net</a></div>')
            button_layout.addWidget(mediainfo_label)
        button_layout.addStretch(1)
        button_layout.addWidget(keyframesButton)
        button_layout.addWidget(okButton)
        layout = QVBoxLayout()
        layout.addWidget(content)
        layout.addLayout(button_layout)
        self.setLayout(layout)

    def showKeyframes(self):
        qApp.setOverrideCursor(Qt.WaitCursor)
        keyframes = self.parent.videoService.getKeyframes(self.media, formatted_time=True)
        kframes = KeyframesDialog(keyframes, self)
        kframes.show()


class KeyframesDialog(QDialog):
    def __init__(self, keyframes: list, parent=None, flags=Qt.Tool | Qt.FramelessWindowHint):
        super(KeyframesDialog, self).__init__(parent, flags)
        self.parent = parent
        self.setObjectName('keyframes')
        self.setAttribute(Qt.WA_DeleteOnClose, True)
        self.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Preferred)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle('View keyframes')
        self.setStyleSheet('QDialog { border: 2px solid #000; }')
        self.content = QTextBrowser(self)
        self.content.setStyleSheet('''
            QTextBrowser {{
                border: none;
                border-top: 1px solid {0};
                border-bottom: 1px solid {0};
                background-color: transparent;
        }}'''.format('#4D5355' if parent.parent.theme == 'dark' else '#C0C2C3'))
        self.content.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        buttons = QDialogButtonBox(QDialogButtonBox.Close)
        buttons.accepted.connect(self.close)
        buttons.rejected.connect(self.close)
        content_headers = '''<style>
                    table td h2 {{
                        font-family: "Futura LT", sans-serif;
                        font-weight: normal;
                        text-align: center;
                        color: {}
                    }}
                </style>
                <table width="300" border="0" cellpadding="4" cellspacing="0">
                    <tr>
                        <td><h2>Keyframe Timecodes</h2></td>
                    </tr>
                </table>
                '''.format('#C681D5' if parent.parent.theme == 'dark' else '#642C68')
        headers = QLabel(content_headers, self)
        halver = math.ceil(len(keyframes) / 2)
        col1 = '<br/>'.join(keyframes[:halver])
        col2 = '<br/>'.join(keyframes[halver:])
        html = '''<style>
            table {{
               font-family: "Noto Sans", sans-serif;
               font-size: 13px;
           }}
           td {{
               text-align: center;
               font-weight: normal;
           }}
        </style>
        <div align="center">
           <table border="0" cellpadding="2" cellspacing="0">
               <tr>
                   <td>{col1}</td>
                   <td>{col2}</td>
               </tr>
           </table>
        </div>'''.format(**locals())
        self.content.setHtml(html)
        totalLabel = QLabel('<b>total keyframes:</b> {}'.format(len(keyframes)), self)
        totalLabel.setAlignment(Qt.AlignCenter)
        totalLabel.setObjectName('modalfooter')
        totalLabel.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)
        button_layout = QHBoxLayout()
        button_layout.addWidget(totalLabel, 1)
        button_layout.addWidget(buttons, 0)
        layout = QVBoxLayout()
        layout.addWidget(headers)
        layout.addWidget(self.content)
        layout.addLayout(button_layout)
        self.setLayout(layout)
        self.setMinimumWidth(320)
        qApp.restoreOverrideCursor()

    def showEvent(self, event: QShowEvent) -> None:
        if self.content.verticalScrollBar().isVisible() \
                and (self.parent.parent.parent.stylename == 'fusion' or sys.platform in {'win32', 'darwin'}):
            self.content.setStyleSheet('''
                QTextBrowser {{
                    border-left: none;
                    border-right: 1px solid {0};
                    border-top: 1px solid {0};
                    border-bottom: 1px solid {0};
                    background-color: transparent;
                }}'''.format('#4D5355' if self.parent.parent.theme == 'dark' else '#C0C2C3'))
        super(KeyframesDialog, self).showEvent(event)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import os
import time

from PyQt5.QtCore import pyqtSignal, pyqtSlot, Qt, QTimer, QUrl
from PyQt5.QtGui import QDesktopServices, QIcon, QMouseEvent
from PyQt5.QtWidgets import qApp, QDialog, QHBoxLayout, QLabel, QPushButton, QVBoxLayout


class Notification(QDialog):
    shown = pyqtSignal()
    closed = pyqtSignal()
    duration = 6

    def __init__(self, icon: str, parent=None):
        super(Notification, self).__init__(parent)
        self.parent = parent
        self.theme = self.parent.theme
        self.setObjectName('notification')
        self.setContentsMargins(10, 10, 10, 10)
        self.setModal(True)
        self.setWindowFlags(Qt.Window | Qt.Dialog | Qt.FramelessWindowHint)
        self.setMinimumWidth(550)
        self.shown.connect(lambda: QTimer.singleShot(self.duration * 1000, self.close))
        self._title, self._message = '', ''
        self.buttons = []
        self.msgLabel = QLabel(self._message, self)
        self.msgLabel.setWordWrap(True)
        logo_label = QLabel('<img src="{}" width="82" />'.format(icon), self)
        logo_label.setFixedSize(82, 82)
        self.left_layout = QVBoxLayout()
        self.left_layout.addWidget(logo_label)
        layout = QHBoxLayout()
        layout.addStretch(1)
        layout.addLayout(self.left_layout)
        layout.addSpacing(10)
        layout.addWidget(self.msgLabel, Qt.AlignVCenter)
        layout.addStretch(1)
        self.setLayout(layout)

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, value):
        self._title = value

    @property
    def message(self):
        return self._message

    @message.setter
    def message(self, value):
        self.msgLabel.setText(value)
        self._message = value

    def mousePressEvent(self, event: QMouseEvent):
        if event.button() == Qt.LeftButton:
            self.close()

    @pyqtSlot()
    def showEvent(self, event):
        if self.isVisible():
            self.shown.emit()
        [self.left_layout.addWidget(btn) for btn in self.buttons]
        super(Notification, self).showEvent(event)

    def closeEvent(self, event):
        event.accept()
        for step in range(100, 0, -10):
            self.setWindowOpacity(step / 100)
            qApp.processEvents()
            time.sleep(0.085)
        self.closed.emit()
        self.deleteLater()
        super(Notification, self).closeEvent(event)


class JobCompleteNotification(Notification):
    def __init__(self, filename: str, filesize: str, runtime: str, icon: str, parent=None):
        super(JobCompleteNotification, self).__init__(icon, parent)
        pencolor = '#C681D5' if self.theme == 'dark' else '#642C68'
        self.filename = filename
        self.filesize = filesize
        self.runtime = runtime
        self.parent = parent
        self.title = 'Your media file is ready!'
        self.message = '''
    <style>
        h1 {{
            color: {labelscolor};
            font-family: "Futura LT", sans-serif;
            font-weight: normal;
            text-align: center;
        }}
        table.info {{
            margin: 6px;
            padding: 4px 2px;
            font-family: "Noto Sans", sans-serif;
        }}
        td.label {{
            font-weight: bold;
            color: {labelscolor};
            text-transform: lowercase;
            text-align: right;
            padding-right: 5px;
            font-size: 14px;
        }}
        td.value {{
            color: {valuescolor};
            font-size: 14px;
        }}
    </style>
    <h1>{heading}</h1>
    <table border="0" class="info" cellpadding="2" cellspacing="0" align="left">
        <tr>
            <td width="20%%" class="label"><b>File:</b></td>
            <td width="80%%" class="value" nowrap>{filename}</td>
        </tr>
        <tr>
            <td width="20%%" class="label"><b>Size:</b></td>
            <td width="80%%" class="value">{filesize}</td>
        </tr>
        <tr>
            <td width="20%%" class="label"><b>Runtime:</b></td>
            <td width="80%%" class="value">{runtime}</td>
        </tr>
    </table>'''.format(labelscolor=pencolor,
                       valuescolor=('#EFF0F1' if self.theme == 'dark' else '#222'),
                       heading=self._title,
                       filename=os.path.basename(self.filename),
                       filesize=self.filesize,
                       runtime=self.runtime)
        playButton = QPushButton(QIcon(':/images/complete-play.png'), 'Play', self)
        playButton.setFixedWidth(82)
        playButton.clicked.connect(self.playMedia)
        playButton.setCursor(Qt.PointingHandCursor)
        self.buttons.append(playButton)

    @pyqtSlot()
    def playMedia(self) -> None:
        if os.path.isfile(self.filename):
            QDesktopServices.openUrl(QUrl.fromLocalFile(self.filename))



# #!/usr/bin/env python3
# -*- coding: utf-8 -*-

#######################################################################
#
# VidCutter - media cutter & joiner
#
# copyright © 2018 Pete Alexandrou
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#######################################################################

import logging
import sys
from datetime import datetime

# PyQt5 5.15.8 changed the import name
try:
    from PyQt5.Qt import PYQT_VERSION_STR
except ImportError:
    from PyQt5.QtCore import PYQT_VERSION_STR
from PyQt5.QtCore import QFile, QObject, QSize, QTextStream, Qt
from PyQt5.QtGui import QPixmap
from PyQt5.QtWidgets import (QDialog, QDialogButtonBox, QHBoxLayout, QLabel, QScrollArea, QSizePolicy, QStyleFactory,
                             QTabWidget, QVBoxLayout, QWidget, qApp)

from vidcutter.libs.config import cached_property

import vidcutter


class About(QDialog):
    def __init__(self, ffmpeg_service: QObject, mpv_service: QObject, parent: QWidget):
        super(About, self).__init__(parent)
        self.parent = parent
        self.logger = logging.getLogger(__name__)
        self.ffmpeg_service = ffmpeg_service
        self.mpv_service = mpv_service
        self.theme = self.parent.theme
        self.setObjectName('aboutwidget')
        self.setContentsMargins(0, 0, 0, 0)
        self.setWindowFlags(Qt.Window | Qt.Dialog | Qt.WindowCloseButtonHint)
        self.setWindowModality(Qt.ApplicationModal)
        logolabel = QLabel(self)
        logolabel.setPixmap(QPixmap(':/images/about-logo.png'))
        versionlabel = QLabel('''
            <style>
                b.label {{
                    font-family: "Noto Sans", sans-serif;
                    font-size: 14px;
                    font-weight: 500;
                    color: {0};
                }}
                b.version {{
                    font-family: "Futura LT", sans-serif;
                    font-size: 16px;
                    font-weight: 600;
                }}
            </style>
            <b class="label">version:</b>&nbsp; <b class="version">{1}</b>
        '''.format('#EA95FF' if self.theme == 'dark' else '#441D4E', qApp.applicationVersion(), self))
        if self.parent.parent.flatpak:
            versionlabel.setText(versionlabel.text() + ' <span style="font-size:12px;">- FLATPAK</span>')
            versionspacing = 75
        else:
            versionspacing = 95
        versionlabel.setAlignment(Qt.AlignBottom)
        versionlayout = QHBoxLayout()
        versionlayout.setSpacing(0)
        versionlayout.setContentsMargins(0, 0, 0, 0)
        versionlayout.addStretch(1)
        versionlayout.addWidget(versionlabel)
        versionlayout.addSpacing(versionspacing)
        headerlayout = QVBoxLayout()
        headerlayout.setSpacing(0)
        headerlayout.setContentsMargins(0, 0, 0, 0)
        headerlayout.addWidget(logolabel, 1)
        headerlayout.addLayout(versionlayout)
        self.tab_about = AboutTab(self)
        self.tab_credits = CreditsTab(self)
        self.tab_license = LicenseTab(self)
        scrollarea = QScrollArea(self)
        scrollarea.setStyleSheet('QScrollArea { background:transparent; }')
        scrollarea.setWidgetResizable(True)
        scrollarea.setHorizontalScrollBarPolicy(Qt.ScrollBarAlwaysOff)
        scrollarea.setFrameShape(QScrollArea.NoFrame)
        scrollarea.setWidget(self.tab_license)
        if sys.platform in {'win32', 'darwin'}:
            scrollarea.setStyle(QStyleFactory.create('Fusion'))
        tabs = QTabWidget()
        tabs.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        tabs.addTab(self.tab_about, 'About')
        tabs.addTab(self.tab_credits, 'Credits')
        tabs.addTab(scrollarea, 'License')
        buttons = QDialogButtonBox(QDialogButtonBox.Ok)
        buttons.accepted.connect(self.close)
        layout = QVBoxLayout()
        layout.addLayout(headerlayout)
        layout.addWidget(tabs, 1)
        layout.addWidget(buttons)
        self.setLayout(layout)
        self.setWindowTitle('About {}'.format(qApp.applicationName()))
        self.setFixedSize(self.sizeHint())

    def sizeHint(self) -> QSize:
        modes = {
            'LOW': QSize(450, 300),
            'NORMAL': QSize(500, 510 if sys.platform == 'darwin' else 490),
            'HIGH': QSize(1080, 920)
        }
        return modes[self.parent.parentWidget().scale]


class BaseTab(QLabel):
    def __init__(self, parent=None):
        super(BaseTab, self).__init__(parent)
        self.setTextFormat(Qt.RichText)
        self.setWordWrap(True)
        self.setAlignment(Qt.AlignLeft | Qt.AlignTop)
        self.setOpenExternalLinks(True)
        if parent.theme == 'dark':
            bgcolor = 'rgba(12, 15, 16, 210)'
            pencolor = '#FFF'
        else:
            bgcolor = 'rgba(255, 255, 255, 200)'    
            pencolor = '#000'
        self.setStyleSheet('''
            QLabel {{
                background-color: {bgcolor};
                color: {pencolor};
                padding: 8px;
            }}'''.format(**locals()))


# noinspection PyBroadException
class AboutTab(BaseTab):
    def __init__(self, parent):
        super(AboutTab, self).__init__(parent)
        self.parent = parent
        self.missing = '<span style="color:maroon; font-weight:bold;">MISSING</span>'
        pencolor = '#FAFAFA' if self.parent.theme == 'dark' else '#222'
        linkcolor = '#EA95FF' if self.parent.theme == 'dark' else '#441D4E'
        mpv_version = self.mpv_version
        ffmpeg_version = self.ffmpeg_version
        python_version = sys.version.split(' ')[0]
        pyqt_version = PYQT_VERSION_STR
        year = datetime.now().year
        mailto = vidcutter.__email__
        author = vidcutter.__author__
        website = vidcutter.__website__
        bugreport = vidcutter.__bugreport__
        self.setText('''
<style>
    table {{ width:100%; font-family:"Noto Sans", sans-serif; background-color:transparent; }}
    td.label {{ font-size:13px; font-weight:bold; text-align:right; }}
    td.value {{
        color: {pencolor};
        font-weight: 600;
        font-family: "Futura LT", sans-serif;
        font-size: 12.5px;
        vertical-align: bottom;
    }}
    a {{ color: {linkcolor}; text-decoration:none; font-weight:bold; }}
</style>
<table border="0" cellpadding="0" cellspacing="4">
    <tr>
        <td>
            <table border="0" cellpadding="0" cellspacing="0">
                <tr valign="top">
                    <td>
                        <table cellpadding="2" cellspacing="0" border="0">
                            <tr>
                                <td class="label">libmpv:</td>
                                <td class="value">{mpv_version}</td>
                                <td width="35" rowspan="2">&nbsp;</td>
                                <td class="label">PyQt:</td>
                                <td class="value">{pyqt_version}</td>
                            </tr>
                            <tr>
                                <td class="label">FFmpeg:</td>
                                <td class="value">{ffmpeg_version}</td>
                                <td class="label">Python:</td>
                                <td class="value">{python_version}</td>
                            </tr>
                        </table>
                        <p style="font-size:13px;">
                            <img src=":/images/copyright.png" style="vertical-align:bottom;" />
                            &nbsp;
                            Copyright {year} <a href="mailto:{mailto}">{author}</a>
                            <br/>
                            <img src=":/images/home.png" style="vertical-align:bottom;" />
                            &nbsp;
                            <a href="{website}">{website}</a>
                        </p>
                        <p style="font-size:13px;">
                            Found a bug? You can <a href="{bugreport}">REPORT IT HERE</a>.
                        </p>
                    </td>
                    <td align="right" nowrap style="font-weight:500;font-size:13px;">
                        <p><b>built using</b></p>
                        <p>
                            <a href="https://python.org" title="Python"><img src=":/images/python.png" /></a>
                            &nbsp;
                            <a href="https://qt.io" title="Qt5"><img src=":/images/qt.png" /></a>
                            <br/><br/>
                            <a href="https://www.jetbrains.com/pycharm" title="PyCharm Professional">
                                <img src=":/images/pycharm.png" />
                            </a>
                        </p>
                    </td>
                </tr>
            </table>
            <p style="font-size:12px;">
                Built in Python with the help of <a href="https://www.jetbrains.com/pycharm">PyCharm Professional</a>
                using an open-source development license donated by its wickedly cool creators at
                <a href="https://www.jetbrains.com">JetBrains</a>.
            </p>
            <p style="font-size:11px; margin-top:15px;">
                This program is free software; you can redistribute it and/or
                modify it under the terms of the GNU General Public License
                version 3, or (at your option) any later version. This software uses code
                produced by the <a href="https://mpv.io">mpv</a> and
                <a href="https://www.ffmpeg.org">FFmpeg</a> projects under the
                <a href="https://www.gnu.org/licenses/old-licenses/gpl-2.0.html">GPLv2.0</a> and
                <a href="https://www.gnu.org/licenses/old-licenses/lgpl-2.1.en.html">LGPLv2.1</a>
                licenses respectively.
            </p>
        </td>
    </tr>
</table>'''.format(**locals()))

    @cached_property
    def mpv_version(self) -> str:
        try:
            return self.parent.mpv_service.version()
        except Exception:
            self.parent.logger.exception('mpv version error', exc_info=True)
            return self.missing

    @cached_property
    def ffmpeg_version(self) -> str:
        try:
            v = self.parent.ffmpeg_service.version()
            return '-'.join(v.replace('~', '-').split('-')[0:2])
        except Exception:
            self.parent.logger.exception('ffmpeg version error', exc_info=True)
            return self.missing


class CreditsTab(BaseTab):
    def __init__(self, parent):
        super(CreditsTab, self).__init__(parent)
        self.parent = parent
        self.setObjectName('credits')
        self.setText('''
        <style>
            table { background-color: transparent; }
            a { color:%s; text-decoration:none; font-weight:bold; }
        </style>
        <h3 style="text-align:center;">CREDITS</h3>
        <p>
            This application either uses code and tools from the following projects in part or in their entirety as
            deemed permissable by each project's open-source license.
        </p>
        <table border="0" cellpadding="10" cellspacing="0" width="400" align="center" style="margin-top:10px;">
            <tr>
                <td width="200">
                    <p>
                        <a href="https://github.com/marcan/pympv">pympv</a>
                        -
                        GPLv3+
                    </p>
                    <p>
                        <a href="http://ffmpeg.org">FFmpeg</a>
                        -
                        GPLv2+
                    </p>
                    <p>
                        <a href="https://www.riverbankcomputing.com/software/pyqt">PyQt5</a>
                        -
                        GPLv3+
                    </p>
                </td>
                <td width="200">
                    <p>
                        <a href="http://mpv.io">libmpv</a>
                        -
                        GPLv3+
                    </p>
                    <p>
                        <a href="http://mediaarea.net/mediainfo">MediaInfo</a>
                        -
                        BSD-style
                    </p>
                    <p>
                        <a href="https://www.qt.io">Qt5</a>
                        -
                        LGPLv3
                    </p>
                </td>
            </tr>
        </table>
        <p>
            Click on a project name to visit the official website.
        </p>''' % ('#EA95FF' if self.parent.theme == 'dark' else '#441D4E'))


class LicenseTab(BaseTab):
    def __init__(self, parent):
        super(LicenseTab, self).__init__(parent)
        self.setObjectName('license')
        licensefile = QFile(':/license.html')
        licensefile.open(QFile.ReadOnly | QFile.Text)
        content = QTextStream(licensefile).readAll()
        self.setText(content)
        if sys.platform in {'win32', 'darwin'}:
            self.setStyle(QStyleFactory.create('Fusion'))
